from warnings import warn

import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.mixture import BayesianGaussianMixture, GaussianMixture
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted
from sklearn_compat.utils.validation import validate_data


class GaussianMixtureNB(ClassifierMixin, BaseEstimator):
    """The `GaussianMixtureNB` estimator is a naive bayes classifier that uses a mixture of gaussians instead of
    merely a single one. In particular it trains a `GaussianMixture` model for each class in the target and for each
    feature in the data, on the subset of `X` where `y == class`.

    You can pass any keyword parameter that scikit-learn's
    [GaussianMixture](https://scikit-learn.org/stable/modules/generated/sklearn.mixture.GaussianMixture.html)
    model uses and it will be passed along to each of the models.

    Attributes
    ----------
    gmms_ : dict[int, List[GaussianMixture]]
        A dictionary of Gaussian Mixture Models, one for each class.
    classes_ : np.ndarray of shape (n_classes,)
        The classes seen during `fit`.
    n_features_in_ : int
        The number of features seen during `fit`.
    num_fit_cols_ : int
        Deprecated, please use `n_features_in_` instead.
    """

    def __init__(
        self,
        n_components=1,
        covariance_type="full",
        tol=1e-3,
        reg_covar=1e-6,
        max_iter=100,
        n_init=1,
        init_params="kmeans",
        weights_init=None,
        means_init=None,
        precisions_init=None,
        random_state=None,
        warm_start=False,
    ):
        self.n_components = n_components
        self.covariance_type = covariance_type
        self.tol = tol
        self.reg_covar = reg_covar
        self.max_iter = max_iter
        self.n_init = n_init
        self.init_params = init_params
        self.weights_init = weights_init
        self.means_init = means_init
        self.precisions_init = precisions_init
        self.random_state = random_state
        self.warm_start = warm_start

    def fit(self, X, y) -> "GaussianMixtureNB":
        """Fit the `GaussianMixtureNB` estimator using `X` and `y` as training data by fitting a `GaussianMixture` model
        for each class in the target and for each feature in the data, on the subset of `X` where `y == class`.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features )
            The training data.
        y : array-like of shape (n_samples,)
            The target values.

        Returns
        -------
        self : GaussianMixtureNB
            The fitted estimator.
        """
        X, y = validate_data(self, X=X, y=y, dtype=FLOAT_DTYPES, reset=True)
        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        self.gmms_ = {}
        self.classes_ = unique_labels(y)
        self.n_features_in_ = X.shape[1]
        for c in self.classes_:
            subset_x, subset_y = X[y == c], y[y == c]
            self.gmms_[c] = [
                GaussianMixture(
                    n_components=self.n_components,
                    covariance_type=self.covariance_type,
                    tol=self.tol,
                    reg_covar=self.reg_covar,
                    max_iter=self.max_iter,
                    n_init=self.n_init,
                    init_params=self.init_params,
                    weights_init=self.weights_init,
                    means_init=self.means_init,
                    precisions_init=self.precisions_init,
                    random_state=self.random_state,
                    warm_start=self.warm_start,
                ).fit(subset_x[:, i].reshape(-1, 1), subset_y)
                for i in range(X.shape[1])
            ]
        self.n_iter_ = sum(sum(gmm.n_iter_ for gmm in gmm_c) for gmm_c in self.gmms_.values())
        return self

    def predict(self, X):
        """Predict labels for `X` using fitted estimator and `predict_proba` method, by picking the class with the
        highest probability.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples,)
            The predicted data.
        """
        check_is_fitted(self, ["gmms_", "classes_", "n_features_in_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)
        # if self.n_features_in_ != X.shape[1]:
        #     raise ValueError(f"number of columns {X.shape[1]} does not match fit size {self.n_features_in_}")

        return self.classes_[self.predict_proba(X).argmax(axis=1)]

    def predict_proba(self, X: np.ndarray):
        """Predict probabilities for `X` using fitted estimator by summing the probabilities of each gaussian for each
        class.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples, n_classes)
            The predicted probabilities.
        """
        check_is_fitted(self, ["gmms_", "classes_", "n_features_in_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)

        probs = np.zeros((X.shape[0], len(self.classes_)))
        for k, v in self.gmms_.items():
            class_idx = np.argmax(self.classes_ == k)
            probs[:, class_idx] = np.array(
                [m.score_samples(np.expand_dims(X[:, idx], 1)) for idx, m in enumerate(v)]
            ).sum(axis=0)
        likelihood = np.exp(probs)
        return likelihood / likelihood.sum(axis=1).reshape(-1, 1)

    @property
    def num_fit_cols_(self):
        warn(
            "Please use `n_features_in_` instead of `num_fit_cols_`,"
            "`num_fit_cols_` will be deprecated in future versions",
            DeprecationWarning,
        )
        return self.n_features_in_


class BayesianGaussianMixtureNB(ClassifierMixin, BaseEstimator):
    """The `BayesianGaussianMixtureNB` estimator is a naive bayes classifier that uses a bayesian mixture of gaussians
    instead of merely a single one. In particular it trains a `BayesianGaussianMixture` model for each class in the
    target and for each feature in the data, on the subset of `X` where `y == class`.

    You can pass any keyword parameter that scikit-learn's
    [`BayesianGaussianMixture`](https://scikit-learn.org/stable/modules/generated/sklearn.mixture.BayesianGaussianMixture.html)
    model uses and it will be passed along to each of the models.

    Attributes
    ----------
    gmms_ : dict[int, List[BayesianGaussianMixture]]
        A dictionary of Gaussian Mixture Models, one for each class.
    classes_ : np.ndarray of shape (n_classes,)
        The classes seen during `fit`.
    n_features_in_ : int
        The number of features seen during `fit`.
    num_fit_cols_ : int
        Deprecated, please use `n_features_in_` instead.
    """

    def __init__(
        self,
        n_components=1,
        covariance_type="full",
        tol=0.001,
        reg_covar=1e-06,
        max_iter=100,
        n_init=1,
        init_params="kmeans",
        weight_concentration_prior_type="dirichlet_process",
        weight_concentration_prior=None,
        mean_precision_prior=None,
        mean_prior=None,
        degrees_of_freedom_prior=None,
        covariance_prior=None,
        random_state=None,
        warm_start=False,
        verbose=0,
        verbose_interval=10,
    ):
        self.n_components = n_components
        self.covariance_type = covariance_type
        self.tol = tol
        self.reg_covar = reg_covar
        self.max_iter = max_iter
        self.n_init = n_init
        self.init_params = init_params
        self.weight_concentration_prior_type = weight_concentration_prior_type
        self.weight_concentration_prior = weight_concentration_prior
        self.mean_precision_prior = mean_precision_prior
        self.mean_prior = mean_prior
        self.degrees_of_freedom_prior = degrees_of_freedom_prior
        self.covariance_prior = covariance_prior
        self.random_state = random_state
        self.warm_start = warm_start
        self.verbose = verbose
        self.verbose_interval = verbose_interval

    def fit(self, X, y) -> "BayesianGaussianMixtureNB":
        """Fit the `BayesianGaussianMixtureNB` estimator using `X` and `y` as training data by fitting a
        `BayesianGaussianMixture` model for each class in the target and for each feature in the data, on the subset of
        `X` where `y == class`.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features )
            The training data.
        y : array-like of shape (n_samples,)
            The target values.

        Returns
        -------
        self : BayesianGaussianMixtureNB
            The fitted estimator.
        """
        X, y = validate_data(self, X=X, y=y, dtype=FLOAT_DTYPES, reset=True)
        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        self.gmms_ = {}
        self.classes_ = unique_labels(y)
        self.n_features_in_ = X.shape[1]
        for c in self.classes_:
            subset_x, subset_y = X[y == c], y[y == c]
            self.gmms_[c] = [
                BayesianGaussianMixture(
                    n_components=self.n_components,
                    covariance_type=self.covariance_type,
                    tol=self.tol,
                    reg_covar=self.reg_covar,
                    max_iter=self.max_iter,
                    n_init=self.n_init,
                    init_params=self.init_params,
                    weight_concentration_prior_type=self.weight_concentration_prior_type,
                    weight_concentration_prior=self.weight_concentration_prior,
                    mean_precision_prior=self.mean_precision_prior,
                    mean_prior=self.mean_prior,
                    degrees_of_freedom_prior=self.degrees_of_freedom_prior,
                    covariance_prior=self.covariance_prior,
                    random_state=self.random_state,
                    warm_start=self.warm_start,
                    verbose=self.verbose,
                    verbose_interval=self.verbose_interval,
                ).fit(subset_x[:, i].reshape(-1, 1), subset_y)
                for i in range(X.shape[1])
            ]
        self.n_iter_ = sum(sum(gmm.n_iter_ for gmm in gmm_c) for gmm_c in self.gmms_.values())
        return self

    def predict(self, X):
        """Predict labels for `X` using fitted estimator and `predict_proba` method, by picking the class with the
        highest probability.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples,)
            The predicted data.
        """
        check_is_fitted(self, ["gmms_", "classes_", "n_features_in_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)

        return self.classes_[self.predict_proba(X).argmax(axis=1)]

    def predict_proba(self, X: np.ndarray):
        """Predict probabilities for `X` using fitted estimator by summing the probabilities of each gaussian for each
        class.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples, n_classes)
            The predicted probabilities.
        """
        check_is_fitted(self, ["gmms_", "classes_", "n_features_in_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)

        probs = np.zeros((X.shape[0], len(self.classes_)))
        for k, v in self.gmms_.items():
            class_idx = np.argmax(self.classes_ == k)
            probs[:, class_idx] = np.array(
                [m.score_samples(np.expand_dims(X[:, idx], 1)) for idx, m in enumerate(v)]
            ).sum(axis=0)
        likelihood = np.exp(probs)
        return likelihood / likelihood.sum(axis=1).reshape(-1, 1)

    @property
    def num_fit_cols_(self):
        warn(
            "Please use `n_features_in_` instead of `num_fit_cols_`,"
            "`num_fit_cols_` will be deprecated in future versions",
            DeprecationWarning,
        )
        return self.n_features_in_



import sys

import numpy as np
import pandas as pd

if sys.version_info >= (3, 9):
    import importlib.resources as importlib_resources  # pragma: no cover
else:
    import importlib_resources as importlib_resources  # pragma: no cover


from sklearn.datasets import fetch_openml


def load_penguins(return_X_y=False, as_frame=False):
    """Loads the penguins dataset, which is a lovely alternative for the iris dataset.
    We've added this dataset for educational use.

    Data were collected and made available by Dr. Kristen Gorman and the Palmer Station, Antarctica LTER, a member of
    the Long Term Ecological Research Network. The goal of the dataset is to predict which species of penguin a penguin
    belongs to.

    This data originally appeared as a R package and R users can find this data in the
    [palmerpenguins package](https://github.com/allisonhorst/palmerpenguins).
    You can also go to the repository for some lovely images that explain the dataset.

    To cite this dataset in publications use:

        Gorman KB, Williams TD, Fraser WR (2014) Ecological Sexual Dimorphism
        and Environmental Variability within a Community of Antarctic
        Penguins (Genus Pygoscelis). PLoS ONE 9(3): e90081.
        https://doi.org/10.1371/journal.pone.0090081

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns `(data, target)` instead of a dict object.
    as_frame : bool, default=False
        If True, returns `data` as a pandas DataFrame  instead of X, y matrices.

    Examples
    -------
    ```py
    from sklego.datasets import load_penguins
    X, y = load_penguins(return_X_y=True)

    X.shape
    # (344, 6)
    y.shape
    # (344,)

    load_penguins(as_frame=True).columns
    # Index(['species', 'island', 'bill_length_mm', 'bill_depth_mm',
    #    'flipper_length_mm', 'body_mass_g', 'sex'],
    #   dtype='object')
    ```

    Notes
    -----
    Anyone interested in publishing the data should contact
    [`Dr. Kristen Gorman`](https://www.uaf.edu/cfos/people/faculty/detail/kristen-gorman.php)
    about analysis and working together on any final products.

    !!! quote "From Gorman et al. (2014)"

        Data reported here are publicly available within the
        [PAL-LTER data system (datasets 219, 220, and 221)](http://oceaninformatics.ucsd.edu/datazoo/data/pallter/datasets).

        Individuals interested in using these data are therefore expected to follow the [US LTER Network's Data Access
        Policy, Requirements and Use Agreement](https://lternet.edu/data-access-policy/)

    Please cite data using the following
    ------------------------------------
    **Adélie penguins:**

      - [Palmer Station Antarctica LTER and K. Gorman, 2020. Structural size measurements and isotopic signatures of
        foraging among adult male and female Adélie penguins (*Pygoscelis adeliae*) nesting along the Palmer Archipelago
        near Palmer Station, 2007-2009 ver 5. Environmental Data
        Initiative](https://doi.org/10.6073/pasta/98b16d7d563f265cb52372c8ca99e60f).

        (Accessed 2020-06-08).

    **Gentoo penguins:**

      - [Palmer Station Antarctica LTER and K. Gorman, 2020. Structural size measurements and isotopic signatures of
        foraging among adult male and female Gentoo penguin (*Pygoscelis papua*) nesting along the Palmer Archipelago
        near Palmer Station, 2007-2009 ver 5. Environmental Data
        Initiative](https://doi.org/10.6073/pasta/7fca67fb28d56ee2ffa3d9370ebda689).

        (Accessed 2020-06-08).

    **Chinstrap penguins:**

      - [Palmer Station Antarctica LTER and K. Gorman, 2020. Structural size measurements and isotopic signatures of
        foraging among adult male and female Chinstrap penguin (*Pygoscelis antarcticus*) nesting along the Palmer
        Archipelago near Palmer Station, 2007-2009 ver 6.
        Environmental Data Initiative](https://doi.org/10.6073/pasta/c14dfcfada8ea13a17536e73eb6fbe9e).

        (Accessed 2020-06-08).
    """
    filepath = importlib_resources.files("sklego") / "data" / "penguins.zip"
    df = pd.read_csv(filepath)
    if as_frame:
        return df
    X, y = (
        df[
            [
                "island",
                "bill_length_mm",
                "bill_depth_mm",
                "flipper_length_mm",
                "body_mass_g",
                "sex",
            ]
        ].to_numpy(),
        df["species"].to_numpy(),
    )
    if return_X_y:
        return X, y
    return {"data": X, "target": y}


def load_arrests(return_X_y=False, as_frame=False):
    """Loads the arrests dataset which can serve as a benchmark for fairness.

    It is data on the police treatment of individuals arrested in Toronto for simple possession of small quantities of
    marijuana. The goal is to predict whether or not the arrestee was released with a summons while maintaining a
    degree of fairness.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns `(data, target)` instead of a dict object.
    as_frame : bool, default=False
        If True, returns `data` as a pandas DataFrame  instead of X, y matrices.

    Examples
    -------
    ```py
    from sklego.datasets import load_arrests
    X, y = load_arrests(return_X_y=True)

    X.shape
    # (5226, 7)
    y.shape
    # (5226,)

    load_arrests(as_frame=True).columns
    # Index(['released', 'colour', 'year', 'age', 'sex', 'employed', 'citizen',
    #   'checks'],
    #  dtype='object')
    ```

    The dataset was copied from the carData R package
    ([dataset documentation](https://vincentarelbundock.github.io/Rdatasets/doc/carData/Arrests.html))
    and can originally be found in:

    - Personal communication from Michael Friendly, York University.
    """
    filepath = importlib_resources.files("sklego") / "data" / "arrests.zip"
    df = pd.read_csv(filepath)
    if as_frame:
        return df
    X, y = (
        df[["colour", "year", "age", "sex", "employed", "citizen", "checks"]].to_numpy(),
        df["released"].to_numpy(),
    )
    if return_X_y:
        return X, y
    return {"data": X, "target": y}


def load_chicken(return_X_y=False, as_frame=False):
    """Loads the chicken dataset.

    The chicken data has 578 rows and 4 columns from an experiment on the effect of diet on early growth of chicks.
    The body weights of the chicks were measured at birth and every second day thereafter until day 20.
    They were also measured on day 21. There were four groups on chicks on different protein diets.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns `(data, target)` instead of a dict object.
    as_frame : bool, default=False
        If True, returns `data` as a pandas DataFrame  instead of X, y matrices.

    Examples
    -------
    ```py
    from sklego.datasets import load_chicken
    X, y = load_chicken(return_X_y=True)

    X.shape
    # (578, 3)
    y.shape
    # (578,)

    load_chicken(as_frame=True).columns
    # Index(['weight', 'time', 'chick', 'diet'], dtype='object')
    ```

    The datasets can be found in the following sources:

    - Crowder, M. and Hand, D. (1990), Analysis of Repeated Measures, Chapman and Hall (example 5.3)
    - Hand, D. and Crowder, M. (1996), Practical Longitudinal Data Analysis, Chapman and Hall (table A.2)
    """
    filepath = importlib_resources.files("sklego") / "data" / "chickweight.zip"
    df = pd.read_csv(filepath)
    if as_frame:
        return df
    X, y = df[["time", "diet", "chick"]].to_numpy(), df["weight"].to_numpy()
    if return_X_y:
        return X, y
    return {"data": X, "target": y}


def load_abalone(return_X_y=False, as_frame=False):
    """
    Loads the abalone dataset where the goal is to predict the gender of the creature.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns `(data, target)` instead of a dict object.
    as_frame : bool, default=False
        If True, returns `data` as a pandas DataFrame  instead of X, y matrices.

    Examples
    ---------
    ```py
    from sklego.datasets import load_abalone
    X, y = load_abalone(return_X_y=True)

    X.shape
    # (4177, 8)
    y.shape
    # (4177,)

    load_abalone(as_frame=True).columns
    # Index(['sex', 'length', 'diameter', 'height', 'whole_weight', 'shucked_weight',
    #        'viscera_weight', 'shell_weight', 'rings'],
    #       dtype='object')
    ```

    The dataset was copied from Kaggle and can originally be found in the following sources:

    - Warwick J Nash, Tracy L Sellers, Simon R Talbot, Andrew J Cawthorn and Wes B Ford (1994)

        "The Population Biology of Abalone (_Haliotis_ species) in Tasmania."

        Sea Fisheries Division, Technical Report No. 48 (ISSN 1034-3288)
    """
    filepath = importlib_resources.files("sklego") / "data" / "abalone.zip"
    df = pd.read_csv(filepath)
    if as_frame:
        return df
    X = df[
        [
            "length",
            "diameter",
            "height",
            "whole_weight",
            "shucked_weight",
            "viscera_weight",
            "shell_weight",
            "rings",
        ]
    ].to_numpy()
    y = df["sex"].to_numpy()
    if return_X_y:
        return X, y
    return {"data": X, "target": y}


def load_heroes(return_X_y=False, as_frame=False):
    """A dataset from the video game [Heroes of the storm](https://heroesofthestorm.blizzard.com/en-us/).

    The goal of the dataset is to predict the attack type. Note that the pandas dataset returns more information.
    This is because we wanted to keep the X simple in the return_X_y case.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns `(data, target)` instead of a dict object.
    as_frame : bool, default=False
        If True, returns `data` as a pandas DataFrame  instead of X, y matrices.

    Examples
    --------
    ```py
    from sklego.datasets import load_heroes
    X, y = load_heroes(return_X_y=True)

    X.shape
    # (84, 2)
    y.shape
    # (84,)

    load_heroes(as_frame=True).columns
    # Index(['name', 'attack_type', 'role', 'health', 'attack', 'attack_spd'], dtype='object')
    ```
    """
    filepath = importlib_resources.files("sklego") / "data" / "heroes.zip"
    df = pd.read_csv(filepath)
    if as_frame:
        return df
    X = df[["health", "attack"]].to_numpy()
    y = df["attack_type"].to_numpy()
    if return_X_y:
        return X, y
    return {"data": X, "target": y}


def load_hearts(return_X_y=False, as_frame=False):
    """Loads the Cleveland Heart Diseases dataset.

    The goal is to predict the presence of a heart disease (target values 1, 2, 3, and 4).
    The data originates from research to heart diseases by four institutions and originally contains 76 attributes.
    Yet, all published experiments refer to using a subset of 13 features and one target.
    This implementation loads the Cleveland dataset of the research which is the only set used by ML researchers
    to this date.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns `(data, target)` instead of a dict object.
    as_frame : bool, default=False
        If True, returns `data` as a pandas DataFrame  instead of X, y matrices.

    Examples
    --------
    ```py
    from sklego.datasets import load_hearts
    X, y = load_hearts(return_X_y=True)

    X.shape
    # (303, 13)
    y.shape
    # (303,)

    load_hearts(as_frame=True).columns
    # Index(['age', 'sex', 'cp', 'trestbps', 'chol', 'fbs', 'restecg', 'thalach',
    #        'exang', 'oldpeak', 'slope', 'ca', 'thal', 'target'],
    #       dtype='object')
    ```

    The dataset can originally be found in the
    [UC Irvine Machine Learning Repository](https://archive.ics.uci.edu/ml/datasets/Heart+Disease)

    The responsible institutions for the entire database are:

    1. Hungarian Institute of Cardiology. Budapest: Andras Janosi, M.D.
    2. University Hospital, Zurich, Switzerland: William Steinbrunn, M.D.
    3. University Hospital, Basel, Switzerland: Matthias Pfisterer, M.D.
    4. V.A. Medical Center, Long Beach and Cleveland Clinic Foundation: Robert Detrano, M.D., Ph.D.

    The documentation of the dataset can be viewed at:
    https://archive.ics.uci.edu/ml/machine-learning-databases/heart-disease/heart-disease.names
    """
    filepath = importlib_resources.files("sklego") / "data" / "hearts.zip"
    df = pd.read_csv(filepath)
    if as_frame:
        return df
    X = df[
        [
            "age",
            "sex",
            "cp",
            "trestbps",
            "chol",
            "fbs",
            "restecg",
            "thalach",
            "exang",
            "oldpeak",
            "slope",
            "ca",
            "thal",
        ]
    ].to_numpy()
    y = df["target"].to_numpy()
    if return_X_y:
        return X, y
    return {"data": X, "target": y}


def make_simpleseries(
    n_samples=365 * 5,
    trend=0.001,
    season_trend=0.001,
    noise=0.5,
    as_frame=False,
    seed=None,
    stack_noise=False,
    start_date=None,
):
    """Generate a very simple timeseries dataset to play with.

    The generator returns a daily time-series with a yearly seasonality, trend, and noise.

    Parameters
    ----------
    n_samples : int, default=365 * 5
        The number of days to simulate the timeseries for.
    trend : float, default=0.001
        The long term trend in the dataset.
    season_trend : float, default=0.001
        The long term trend in the seasonality.
    noise : float, default=0.5
        The noise that is applied to the dataset.
    as_frame : bool, default=False
        Whether to return a pandas dataframe instead of a numpy array.
    seed : int | None, default=None
        The seed value for the randomness.
    stack_noise : bool, default=False
        Set the noise to be stacked by a cumulative sum.
    start_date : str | None, default=None
        Also add a start date (only works if `as_frame`=True).

    Returns
    -------
    np.ndarray | pd.DataFrame
        Timeseries dataset with specified characteristics.

    Examples
    --------
    ```py
    from sklego.datasets import make_simpleseries

    make_simpleseries(seed=42)
    # array([-0.34078806, -0.61828731, -0.18458236, ..., -0.27547402,
    #        -0.38237413,  0.13489355])

    make_simpleseries(as_frame=True, start_date="2018-01-01", seed=42).head(3)
    '''
             yt       date
    0 -0.340788 2018-01-01
    1 -0.618287 2018-01-02
    2 -0.184582 2018-01-03
    '''
    ```
    """
    if seed:
        np.random.seed(seed)
    time = np.arange(0, n_samples)
    noise = np.random.normal(0, noise, n_samples)
    if stack_noise:
        noise = noise.cumsum()
    r1, r2 = np.random.normal(0, 1, 2)
    seasonality = r1 * np.sin(time / 365 * 2 * np.pi) + r2 * np.cos(time / 365 * 4 * np.pi + 1)
    result = seasonality + season_trend * seasonality * time + trend * time + noise
    if as_frame:
        if start_date:
            stamps = pd.date_range(start_date, periods=n_samples)
            return pd.DataFrame({"yt": result, "date": stamps})
        return pd.DataFrame({"yt": result})
    return result


def fetch_creditcard(*, cache=True, data_home=None, as_frame=False, return_X_y=False):
    """Load the creditcard dataset. Download it if necessary.

    Note that internally this is using
    [`fetch_openml`](https://scikit-learn.org/stable/modules/generated/sklearn.datasets.fetch_openml.html) from
    scikit-learn, which is experimental.

        ==============   ==============
        Samples total            284807
        Dimensionality               29
        Features                   real
        Target                 int 0, 1
        ==============   ==============

    The datasets contains transactions made by credit cards in September 2013 by european cardholders.
    This dataset present transactions that occurred in two days, where we have 492 frauds out of 284,807 transactions.
    The dataset is highly unbalanced, the positive class (frauds) account for 0.172% of all transactions.

    Please cite:

        Andrea Dal Pozzolo, Olivier Caelen, Reid A. Johnson and Gianluca Bontempi.
        Calibrating Probability with Undersampling for Unbalanced Classification.
        In Symposium on Computational Intelligence and Data Mining (CIDM), IEEE, 2015

    Parameters
    ----------
    cache : bool, default=True
        Whether to cache downloaded datasets using joblib.
    data_home : str | None, default=None
        Specify another download and cache folder for the datasets. By default all scikit-learn data is stored in
        '~/scikit_learn_data' subfolders.
    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with appropriate dtypes (numeric, string or
        categorical). The target is a pandas DataFrame or Series depending on the number of target_columns.
        The Bunch will contain a `frame` attribute with the target and the data. If `return_X_y` is True, then
        `(data, target)` will be pandas DataFrames or Series as describe above.
    return_X_y : bool, default=False.
        If True, returns `(data.data, data.target)` instead of a Bunch object.

    Returns
    -------
    Dictionary-like
        With the following attributes:

         * data
            ndarray, shape (284807, 29) if ``as_frame`` is True, ``data`` is a pandas object.
         * target
            ndarray, shape (284807, ) if ``as_frame`` is True, ``target`` is a pandas object.
         * feature_names
            Array of ordered feature names used in the dataset.
         * DESCR
            Description of the creditcard dataset. Best to use print.

    Notes
    -----
    This dataset consists of 284807 samples and 29 features.
    """
    return fetch_openml(
        data_id=1597,
        data_home=data_home,
        cache=cache,
        as_frame=as_frame,
        return_X_y=return_X_y,
    )



from sklearn import clone
from sklearn.base import BaseEstimator, ClassifierMixin, MetaEstimatorMixin
from sklearn.metrics import confusion_matrix
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted
from sklearn_compat.utils.validation import validate_data

from sklego.base import ProbabilisticClassifier


class ConfusionBalancer(ClassifierMixin, MetaEstimatorMixin, BaseEstimator):
    r"""The `ConfusionBalancer` estimator attempts to give its child estimator a more balanced output by learning from
    the confusion matrix during training.

    The idea is that the confusion matrix calculates $P(C_i | M_i)$ where $C_i$ is the actual class and $M_i$ is the
    class that the underlying model gives. We use these probabilities to attempt a more balanced prediction by averaging
    the correction from the confusion matrix with the original probabilities.

    $$P(\text{class}_j) = \alpha P(\text{model}_j) + (1-\alpha) P(\text{class}_j | \text{model}_j) P(\text{model}_j)$$

    Parameters
    ----------
    estimator : scikit-learn compatible classifier
        The estimator to be wrapped, it must implement a `predict_proba` method.
    alpha : float, default=0.5
        Hyperparameter which determines how much smoothing to apply. Must be between 0 and 1.
    cfm_smooth : float, default=0
        Smoothing parameter for the confusion matrices to ensure zeros don't exist.

    Attributes
    ----------
    classes_ : array-like of shape (n_classes,)
        The target class labels.
    cfm_ : array-like of shape (n_classes, n_classes)
        The confusion matrix used for the correction.

     Example
     -------
    ```py
    import numpy as np
    from sklearn.linear_model import LogisticRegression

    from sklego.meta import ConfusionBalancer

    np.random.seed(0)
    n1, n2 = 50, 100
    X = np.concatenate([np.random.normal(0, 1, (n1, 2)), np.random.normal(2, 1, (n2, 2))], axis=0)
    y = np.concatenate([np.zeros((n1, 1)), np.ones((n2, 1))], axis=0).reshape(-1)

    confusion_balancer = ConfusionBalancer(
        estimator=LogisticRegression(),
        alpha=.5,
        cfm_smooth= 1
        )

    # Fit the model
    confusion_balancer.fit(X, y)

    # Predict out of sample datapoint
    dp = np.random.normal(0, 1, (1, 2))
    probs = confusion_balancer.predict_proba(dp) # Get probabilities per class
    preds = confusion_balancer.predict(dp) # Get most likely class

    # print(f'Out of sample datapoint {dp} is predicted to belong in class {int(preds[0])} (Probabilities are {probs} for classes 0 and 1, respectively)')

    ### Out of sample datapoint [[-1.30652685  1.65813068]] is predicted to belong in class 0 (Probabilities are [[0.92173898 0.07826102]] for classes 0 and 1, respectively)
    ```
    """

    _required_parameters = ["estimator"]

    def __init__(self, estimator, alpha: float = 0.5, cfm_smooth=0):
        self.estimator = estimator
        self.alpha = alpha
        self.cfm_smooth = cfm_smooth

    def fit(self, X, y):
        """Fit the underlying estimator on the training data `X` and `y`, it calculates the confusion matrix,
        normalizes it and stores it for later use.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        self : ConfusionBalancer
            The fitted estimator.

        Raises
        ------
        ValueError
            If the underlying estimator does not have a `predict_proba` method.
        """

        X, y = validate_data(self, X=X, y=y, dtype=FLOAT_DTYPES, reset=True)

        if not isinstance(self.estimator, ProbabilisticClassifier):
            raise ValueError(
                "The ConfusionBalancer meta model only works on classification models with .predict_proba."
            )
        self.estimator_ = clone(self.estimator).fit(X, y)
        self.classes_ = unique_labels(y)
        cfm = confusion_matrix(y, self.estimator_.predict(X)).T + self.cfm_smooth
        self.cfm_ = cfm / cfm.sum(axis=1).reshape(-1, 1)
        return self

    def predict_proba(self, X):
        """Predict probabilities for new data `X` using the underlying estimator and then applying the confusion matrix
        correction.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples, n_classes)
            The predicted values.
        """
        check_is_fitted(self, ["cfm_", "classes_", "estimator_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)
        preds = self.estimator_.predict_proba(X)
        return (1 - self.alpha) * preds + self.alpha * preds @ self.cfm_

    def predict(self, X):
        """Predict most likely class for new data `X` using the underlying estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples,)
            The predicted values.
        """
        check_is_fitted(self, ["cfm_", "classes_", "estimator_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)
        return self.classes_[self.predict_proba(X).argmax(axis=1)]



import itertools as it
from copy import copy

from sklearn.datasets import make_classification, make_regression
from sklearn.utils._testing import ignore_warnings


@ignore_warnings(category=(DeprecationWarning, FutureWarning))
def check_shape_remains_same_regressor(name, regressor):
    """Ensure that the estimator does not change the shape of prediction."""
    seed = 42

    for n_samples, n_feat in it.product((100, 1000, 10000), (2, 5, 10)):
        regr = copy(regressor)
        X, y = make_regression(
            n_samples=n_samples,
            n_features=n_feat,
            n_informative=n_feat,
            noise=2,
            random_state=seed,
        )
        pred = regr.fit(X, y).predict(X)
        assert y.shape[0] == pred.shape[0]


@ignore_warnings(category=(DeprecationWarning, FutureWarning))
def check_shape_remains_same_classifier(name, classifier):
    """Ensure that the estimator does not change the shape of prediction."""
    seed = 42

    for n_samples, n_feat in it.product((100, 1000, 10000), (2, 5, 10)):
        clf = copy(classifier)
        X, y = make_classification(
            n_samples=n_samples,
            n_features=n_feat,
            n_informative=n_feat,
            n_redundant=0,
            n_repeated=0,
            random_state=seed,
        )
        pred = clf.fit(X, y).predict(X)
        assert y.shape[0] == pred.shape[0]



from typing import List, Union

import narwhals.stable.v1 as nw
import numpy as np
from sklearn.base import BaseEstimator, MetaEstimatorMixin, TransformerMixin, clone
from sklearn.utils.validation import check_is_fitted

from sklego.common import as_list
from sklego.meta._grouped_utils import parse_X_y


class GroupedTransformer(TransformerMixin, MetaEstimatorMixin, BaseEstimator):
    """Construct a transformer per data group. Splits data by groups from single or multiple columns and transforms
    remaining columns using the transformers corresponding to the groups.

    Parameters
    ----------
    transformer : scikit-learn compatible transformer
        The transformer to be applied per group.
    groups : int | str | List[int] | List[str] | None
        The column(s) of the array/dataframe to select as a grouping parameter set. If `None`, the transformer will be
        applied to the entire input without grouping.
    use_global_model : bool, default=True
        Whether or not to fall back to a general transformation in case a group is not found during `.transform()`.
    check_X : bool, default=True
        Whether or not to check the input data. If False, the checks are delegated to the wrapped estimator.

    Attributes
    ----------
    transformers_ : scikit-learn compatible transformer | dict[..., scikit-learn compatible transformer]
        The fitted transformers per group or a single fitted transformer if `groups` is `None`.
    fallback_ : scikit-learn compatible transformer | None
        The fitted transformer to fall back to in case a group is not found during `.transform()`. Only present if
        `use_global_model` is `True`.

    Example
    -------
    ```py
    import pandas as pd
    from sklearn.preprocessing import MinMaxScaler
    from sklego.meta import GroupedTransformer

    results_df = pd.DataFrame(
        {
            "Grade": ["11", "11", "11", "11", "11", "11", "12", "12", "12", "12", "12", "12"],
            "Course": ["Algebra", "Algebra", "Algebra", "English", "English", "English","Algebra", "Algebra", "Algebra", "English", "English", "English"],
            "Name": ["Mary", "Helen", "John", "Mary", "Helen", "John", "Mary", "Helen", "John", "Mary", "Helen", "John"],
            "Result": [100, 94, 97, 88, 92, 96, 97, 98, 96, 90, 92, 94],
        }
    )

    groups = ["Grade", "Course"]
    target = ["Result"]

    # We will use the MinMaxScaler() to scale each grouping (result of course per grade)
    grouped_transformer = GroupedTransformer(MinMaxScaler(), groups)
    grouped_transformer.fit(results_df[groups+target])

    # Scales the result of each student per grade/course
    results_df["Scaled_Result"] = grouped_transformer.transform(results_df[groups+target])
    print(results_df)

    ###   Grade   Course   Name  Result  Scaled_Result
    ###    0     11  Algebra   Mary     100            1.0
    ###    1     11  Algebra  Helen      94            0.0
    ###    2     11  Algebra   John      97            0.5
    ###    3     11  English   Mary      88            0.0
    ###    4     11  English  Helen      92            0.5
    ###    5     11  English   John      96            1.0
    ###    6     12  Algebra   Mary      97            0.5
    ###    7     12  Algebra  Helen      98            1.0
    ###    8     12  Algebra   John      96            0.0
    ###    9     12  English   Mary      90            0.0
    ###    10    12  English  Helen      92            0.5
    ###    11    12  English   John      94            1.0
    ```
    """

    _check_kwargs = {"accept_large_sparse": False}
    _required_parameters = ["transformer", "groups"]

    def __init__(self, transformer, groups, use_global_model=True, check_X=True):
        self.transformer = transformer
        self.groups = groups
        self.use_global_model = use_global_model
        self.check_X = check_X

    def __fit_single_group(self, group, X, y=None):
        """Fit transformer to the given group.

        Parameters
        ----------
        group : tuple
            The group to fit the transformer to.
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,), default=None
            Target values.

        Returns
        -------
        transformer : scikit-learn compatible transformer
            The fitted transformer for the group.
        """
        try:
            return clone(self.transformer).fit(X, y)
        except Exception as e:
            raise type(e)(f"Exception for group {group}: {e}")

    def __fit_grouped_transformer(self, frame: nw.DataFrame, y: Union[np.ndarray, None]):
        """Fit a transformer to each group"""

        grouped_transformers = {
            # Fit a clone of the transformer to each group
            group_name: self.__fit_single_group(
                group_name,
                X=nw.to_native(X_grp.drop(["__sklego_target__", *self.groups_])),
                y=(nw.to_native(X_grp["__sklego_target__"]) if y is not None else None),
            )
            for group_name, X_grp in frame.group_by(self.groups_)
        }

        return grouped_transformers

    def __check_transformer(self):
        """Check if the supplied transformer has a `transform` method and raise a `ValueError` if not."""
        if not hasattr(self.transformer, "transform"):
            raise ValueError("The supplied transformer should have a 'transform' method")

    def fit(self, X, y=None):
        """Fit one transformer for each group of training data `X`.

        Will also learn the groups that exist within the dataset.

        If `use_global_model=True` a fallback transformer will be fitted on the entire dataset in case a group is not
        found during `.transform()`.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data. If `groups` is not `None`, X should have at least two columns, of which at least one
            corresponds to groups defined in `groups`, and the remaining columns represent the values to transform.
        y : array-like of shape (n_samples,), default=None
            Target values.

        Returns
        -------
        self : GroupedTransformer
            The fitted transformer.
        """
        self.__check_transformer()
        self.fallback_ = None
        self.groups_ = as_list(self.groups) if self.groups is not None else []

        X = nw.from_native(X, strict=False, eager_only=True)
        self.n_features_in_ = X.shape[1]

        if isinstance(X, nw.DataFrame):
            self.feature_names_out_ = [c for c in X.columns if c not in self.groups_]

        else:
            # Accounts for negative indices if X is an array
            self.groups_ = [
                X.shape[1] + group if isinstance(group, int) and group < 0 else group for group in self.groups_
            ]
            self.feature_names_out_ = [f"x{i}" for i in range(X.shape[1] - len(self.groups_))]

        frame = parse_X_y(X, y, self.groups_, check_X=self.check_X, **self._check_kwargs)

        if self.groups is None:
            X_, y_ = (
                nw.to_native(frame.drop("__sklego_target__")),
                nw.to_native(frame["__sklego_target__"]) if y is not None else None,
            )
            self.transformers_ = clone(self.transformer).fit(X_, y=y_)
            return self

        self.transformers_ = self.__fit_grouped_transformer(frame, y)

        if self.use_global_model:
            X_, y_ = (
                nw.to_native(frame.drop(["__sklego_target__", *self.groups_])),
                nw.to_native(frame["__sklego_target__"]) if y is not None else None,
            )
            self.fallback_ = clone(self.transformer).fit(X_, y_)

        self.n_features_in_ = X.shape[1]
        return self

    def __transform_single_group(self, group, X):
        """Transform a single group by getting its transformer from the fitted dict"""
        try:
            group_transformer = self.transformers_[group]
        except KeyError:
            if self.fallback_:
                group_transformer = self.fallback_
            else:
                raise ValueError(f"Found new group {group} during transform with use_global_model = False")

        return np.asarray(group_transformer.transform(X))

    def __transform_groups(self, frame: nw.DataFrame):
        """Transform all groups"""

        n_samples = frame.shape[0]
        frame = frame.with_columns(__sklego_index__=np.arange(n_samples))

        results = [
            (
                X_grp.select("__sklego_index__").to_numpy().squeeze().astype(int),
                self.__transform_single_group(
                    group_name, nw.to_native(X_grp.drop(["__sklego_index__", *self.groups_]))
                ),
            )
            for group_name, X_grp in frame.group_by(self.groups_)
        ]

        output = np.empty(shape=(n_samples, results[0][1].shape[1]), dtype=results[0][1].dtype)
        for grp_index, grp_result in results:
            output[grp_index, :] = grp_result

        return output

    def transform(self, X):
        """Transform new data `X` by transforming on each group. If a group is not found during `.transform()` and
        `use_global_model=True` the fallback transformer will be used. If `use_global_model=False` a `ValueError` will
        be raised.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data to transform.

        Returns
        -------
        array-like of shape (n_samples, n_features)
            Data transformed per group.
        """
        check_is_fitted(self, ["n_features_in_", "transformers_"])

        X = nw.from_native(X, strict=False, eager_only=True)
        if X.shape[1] != self.n_features_in_:
            raise ValueError(f"X has {X.shape[1]} features, expected {self.n_features_in_} features.")

        frame = parse_X_y(X, y=None, groups=self.groups_, check_X=self.check_X, **self._check_kwargs).drop(
            "__sklego_target__"
        )

        if self.groups is None:
            X_ = nw.to_native(frame)
            return self.transformers_.transform(X_)

        return self.__transform_groups(frame)

    def _more_tags(self):
        return {"allow_nan": True}

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        return tags

    def get_feature_names_out(self) -> List[str]:
        "Alias for the `feature_names_out_` attribute defined during fit."
        return self.feature_names_out_



from sklearn import clone
from sklearn.base import BaseEstimator, MetaEstimatorMixin, TransformerMixin
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted
from sklearn_compat.utils.validation import _check_n_features, validate_data


class EstimatorTransformer(TransformerMixin, MetaEstimatorMixin, BaseEstimator):
    """Allow using an estimator as a transformer in an earlier step of a pipeline.

    !!! warning
        By default all the checks on the inputs `X` and `y` are delegated to the wrapped estimator.

        To change such behaviour, set `check_input` to `True`.

    Parameters
    ----------
    estimator : scikit-learn compatible estimator
        The estimator to be applied to the data, used as transformer.
    predict_func : str, default="predict"
        The method called on the estimator when transforming e.g. (`"predict"`, `"predict_proba"`).
    check_input : bool, default=False
        Whether or not to check the input data. If False, the checks are delegated to the wrapped estimator.

    Attributes
    ----------
    estimator_ : scikit-learn compatible estimator
        The fitted underlying estimator.
    multi_output_ : bool
        Whether or not the estimator is multi output.

    Example
    -------
    ```py
    import numpy as np
    from sklearn.linear_model import LogisticRegression
    from sklego.meta import EstimatorTransformer

    np.random.seed(0)
    n1, n2 = 50, 100
    X = np.concatenate([np.random.normal(0, 1, (n1, 2)), np.random.normal(2, 1, (n2, 2))], axis=0)
    y = np.concatenate([np.zeros((n1, 1)), np.ones((n2, 1))], axis=0).reshape(-1)

    estimator_transformer = EstimatorTransformer(
        estimator=LogisticRegression(max_iter=1000, random_state=0),
        predict_func="predict_proba",
        check_input= False
        )

    # Fit the data
    estimator_transformer.fit(X, y)
    print(f'Is the output multimodal? {estimator_transformer.multi_output_}')
    ### Is the output multimodal? False

    # Transform the data using "predict_func" as specified
    transformed_X = estimator_transformer.transform(X)
    print(f'Shape of transformed data: {transformed_X.shape}')
    ### Shape of transformed data: (300, 1)
    ```
    """

    def __init__(self, estimator, predict_func="predict", check_input=False):
        self.estimator = estimator
        self.predict_func = predict_func
        self.check_input = check_input

    def fit(self, X, y, **kwargs):
        """Fit the underlying estimator on training data `X` and `y`.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,)
            Target values.
        **kwargs : dict
            Additional keyword arguments passed to the `fit` method of the underlying estimator.

        Returns
        -------
        self : EstimatorTransformer
            The fitted transformer.
        """

        if self.check_input:
            X, y = validate_data(self, X=X, y=y, dtype=FLOAT_DTYPES, multi_output=True, reset=True)
        else:
            _check_n_features(self, X, reset=True)

        self.multi_output_ = len(y.shape) > 1
        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X, y, **kwargs)
        self.n_features_in_ = X.shape[1]
        return self

    def transform(self, X):
        """Transform the data by applying the `predict_func` of the fitted estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to be transformed.

        Returns
        -------
        output : array-like of shape (n_samples,) | (n_samples, n_outputs)
            The transformed data. Array will be of shape `(X.shape[0], )` if estimator is not multi output.
            For multi output estimators an array of shape `(X.shape[0], y.shape[1])` is returned.
        """

        check_is_fitted(self, "estimator_")
        if self.check_input:
            X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)
        else:
            _check_n_features(self, X, reset=False)

        output = getattr(self.estimator_, self.predict_func)(X)
        return output if self.multi_output_ else output.reshape(-1, 1)



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin, clone
from sklearn.linear_model import Ridge
from sklearn.utils.validation import check_is_fitted
from sklearn_compat.utils.validation import _check_n_features, validate_data


class LinearEmbedder(TransformerMixin, BaseEstimator):
    """Embed features using coefficients from a linear model.
    
    The LinearEmbedder fits a linear model to the training data and uses the 
    learned coefficients to rescale the features. This can improve the 
    representation of the features for downstream models, particularly KNN.
    
    Parameters
    ----------
    estimator : sklearn estimator, default=Ridge(fit_intercept=False)
        The linear estimator to use for learning the coefficients.
    check_input : bool, default=True
        Whether to validate input data.
        
    Attributes
    ----------
    estimator_ : sklearn estimator
        The fitted linear estimator.
    coef_ : ndarray of shape (n_features,) or (n_features, n_targets)
        The learned coefficients used for embedding.
    n_features_in_ : int
        Number of features seen during fit.
        
    Examples
    --------
    ```py
    import numpy as np
    from sklearn.datasets import make_regression
    from sklearn.neighbors import KNeighborsRegressor
    from sklearn.pipeline import Pipeline
    from sklego.preprocessing import LinearEmbedder
    
    # Generate sample data
    X, y = make_regression(n_samples=100, n_features=5, noise=0.1, random_state=42)
    
    # Use LinearEmbedder to improve KNN performance
    pipe = Pipeline([
        ('embedder', LinearEmbedder()),
        ('knn', KNeighborsRegressor(n_neighbors=5))
    ])
    
    pipe.fit(X, y)
    predictions = pipe.predict(X)
    ```
    """
    
    def __init__(self, estimator=None, check_input=True):
        self.estimator = estimator
        self.check_input = check_input
        
    def fit(self, X, y):
        """Fit the linear estimator and store the coefficients for embedding.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            Target values.
            
        Returns
        -------
        self : LinearEmbedder
            The fitted embedder.
        """
        if self.check_input:
            X, y = validate_data(self, X=X, y=y, reset=True, multi_output=True)
        else:
            _check_n_features(self, X, reset=True)
            
        # Use Ridge with fit_intercept=False as default
        if self.estimator is None:
            self.estimator_ = Ridge(fit_intercept=False)
        else:
            self.estimator_ = clone(self.estimator)
            
        self.estimator_.fit(X, y)
        
        # Store coefficients for embedding
        self.coef_ = self.estimator_.coef_
        
        # Handle different coefficient shapes
        if len(self.coef_.shape) == 1:
            self.coef_ = self.coef_.reshape(1, -1)
        elif len(self.coef_.shape) == 2 and self.coef_.shape[0] == 1:
            # Keep as is for single output
            pass
        else:
            # For multi-output, take the mean across outputs
            self.coef_ = np.mean(self.coef_, axis=0).reshape(1, -1)
            
        self.n_features_in_ = X.shape[1]
        return self
        
    def transform(self, X):
        """Transform the data by rescaling with learned coefficients.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input data to transform.
            
        Returns
        -------
        X_transformed : ndarray of shape (n_samples, n_features)
            Rescaled data using the learned coefficients.
        """
        check_is_fitted(self, "coef_")
        
        if self.check_input:
            X = validate_data(self, X=X, reset=False)
        else:
            _check_n_features(self, X, reset=False)
            
        # Apply coefficient scaling
        X_transformed = X * self.coef_
        
        return X_transformed


from typing import Callable
from warnings import warn

import numpy as np


def correlation_score(column):
    """The correlation score can score how well the estimator predictions correlate with a given column.

    This is especially useful to use in situations where "fairness" is a theme.

    `correlation_score` takes a column on which to calculate the correlation and returns a metric function.

    Parameters
    ----------
    column : str | int
        Name of the column (when X is a dataframe) or the index of the column (when X is a numpy array) to score
        against.

    Returns
    -------
    Callable[..., float]
        A function which calculates the _negative_ correlation between `estimator.predict(X)` and `X[column]`
        (in gridsearch, larger is better and we want to typically punish correlation).

    Examples
    --------
    ```py
    from sklego.metrics import correlation_score
    ...
    correlation_score('gender')(clf, X, y)
    ```
    """

    def correlation_metric(estimator, X, y_true=None):
        """Remember: X is the thing going *in* to your pipeline."""
        sensitive_col = X[:, column] if isinstance(X, np.ndarray) else X[column]
        return -np.abs(np.corrcoef(estimator.predict(X), sensitive_col)[1, 0])

    return correlation_metric


def p_percent_score(sensitive_column, positive_target=1):
    r"""The p_percent score calculates the ratio between the probability of a positive outcome given the sensitive
    attribute (column) being true and the same probability given the sensitive attribute being false.

    $$\min \left(\frac{P(\hat{y}=1 | z=1)}{P(\hat{y}=1 | z=0)}, \frac{P(\hat{y}=1 | z=0)}{P(\hat{y}=1 | z=1)}\right)$$

    This is especially useful to use in situations where "fairness" is a theme.

    Parameters
    ----------
    sensitive_column : str | int
        Name of the column containing the binary sensitive attribute (when X is a dataframe) or the index of the column
        (when X is a numpy array).
    positive_target : int, default=1
        The name of the class which is associated with a positive outcome

    Returns
    -------
    Callable[..., float]
        A function which calculates the p percent score for z = column

    Examples
    --------
    ```py
    from sklego.metrics import p_percent_score
    ...
    p_percent_score('gender')(clf, X, y)
    ```

    Source
    ------
    M. Zafar et al. (2017), Fairness Constraints: Mechanisms for Fair Classification
    """

    def impl(estimator, X, y_true=None):
        """Remember: X is the thing going *in* to your pipeline."""
        sensitive_col = X[:, sensitive_column] if isinstance(X, np.ndarray) else X[sensitive_column]

        if not ((sensitive_col == 0) | (sensitive_col == 1)).all():
            raise ValueError(
                f"p_percent_score only supports binary indicator columns for `column`. "
                f"Found values {np.unique(sensitive_col)}"
            )

        y_hat = estimator.predict(X)
        y_given_z1 = y_hat[sensitive_col == 1]
        y_given_z0 = y_hat[sensitive_col == 0]
        p_y1_z1 = np.mean(y_given_z1 == positive_target)
        p_y1_z0 = np.mean(y_given_z0 == positive_target)

        # If we never predict a positive target for one of the subgroups, the model is by definition not
        # fair so we return 0
        if p_y1_z1 == 0:
            warn(
                f"No samples with y_hat == {positive_target} for {sensitive_column} == 1, returning 0",
                RuntimeWarning,
            )
            return 0

        if p_y1_z0 == 0:
            warn(
                f"No samples with y_hat == {positive_target} for {sensitive_column} == 0, returning 0",
                RuntimeWarning,
            )
            return 0

        p_percent = np.minimum(p_y1_z1 / p_y1_z0, p_y1_z0 / p_y1_z1)
        return p_percent if not np.isnan(p_percent) else 1

    return impl


def equal_opportunity_score(sensitive_column, positive_target=1):
    r"""The equality opportunity score calculates the ratio between the probability of a **true positive** outcome
    given the sensitive attribute (column) being true and the same probability given the sensitive attribute being
    false.

    $$\min \left(\frac{P(\hat{y}=1 | z=1, y=1)}{P(\hat{y}=1 | z=0, y=1)},
    \frac{P(\hat{y}=1 | z=0, y=1)}{P(\hat{y}=1 | z=1, y=1)}\right)$$

    This is especially useful to use in situations where "fairness" is a theme.

    Parameters
    ----------
    sensitive_column : str | int
        Name of the column containing the binary sensitive attribute (when X is a dataframe) or the index of the column
        (when X is a numpy array).
    positive_target: int, default=1
        The name of the class which is associated with a positive outcome

    Returns
    -------
    Callable[..., float]
        A function which calculates the equal opportunity score for z = column

    Examples
    --------
    ```py
    from sklego.metrics import equal_opportunity_score
    ...
    equal_opportunity_score('gender')(clf, X, y)
    ```

    Source
    ------
    M. Hardt, E. Price and N. Srebro (2016), Equality of Opportunity in Supervised Learning
    """
    def impl(estimator, X, y_true):
        """Remember: X is the thing going *in* to your pipeline."""
        sensitive_col = X[:, sensitive_column] if isinstance(X, np.ndarray) else X[sensitive_column]

        if not ((sensitive_col == 0) | (sensitive_col == 1)).all():
            raise ValueError(
                f"equal_opportunity_score only supports binary indicator columns for `column`. "
                f"Found values {np.unique(sensitive_col)}"
            )

        y_hat = estimator.predict(X)
        y_given_z1_y1 = y_hat[(sensitive_col == 1) & (y_true == positive_target)]
        y_given_z0_y1 = y_hat[(sensitive_col == 0) & (y_true == positive_target)]

        # If we never predict a positive target for one of the subgroups, the model is by definition not
        # fair so we return 0
        if len(y_given_z1_y1) == 0:
            warn(
                f"No samples with y_hat == {positive_target} for {sensitive_column} == 1, returning 0",
                RuntimeWarning,
            )
            return 0

        if len(y_given_z0_y1) == 0:
            warn(
                f"No samples with y_hat == {positive_target} for {sensitive_column} == 0, returning 0",
                RuntimeWarning,
            )
            return 0

        p_y1_z1 = np.mean(y_given_z1_y1 == positive_target)
        p_y1_z0 = np.mean(y_given_z0_y1 == positive_target)
        score = np.minimum(p_y1_z1 / p_y1_z0, p_y1_z0 / p_y1_z1)
        return score if not np.isnan(score) else 1

    return impl


def subset_score(subset_picker: Callable, score: Callable, **kwargs):
    """Return a method that applies the passed score only to a specific subset.

    The subset picker is a method that is passed the corresponding `X` and `y_true` and returns a one-dimensional
    boolean vector where every element corresponds to a row in the data.

    Only the elements with a True value are taken into account for the passed score, representing a filter.

    This allows users to have an easy approach to measuring metrics over different slices of the population which can
    give insights into the model performance, either specifically for fairness or in general.

    Parameters
    ----------
    subset_picker : Callable
        Method that returns a boolean mask that is used for slicing the samples
    score : Callable[..., T]
        The score that needs to be applied to the subset
    kwargs : dict
        Additional keyword arguments to pass to score

    Returns
    -------
    Callable[..., T]
        A function which calculates the passed score for the subset

    Examples
    --------
    ```py
    from sklego.metrics import subset_score
    ...
    subset_score(lambda X, y_true: X['column'] == 'A', accuracy_score)(clf, X, y)
    ```
    """

    def sliced_metric(estimator, X, y_true=None):
        mask = subset_picker(X, y_true)
        if isinstance(mask, np.ndarray):
            if len(mask.shape) > 1:
                raise ValueError(
                    "`subset_picker` should return 1-dimensional numpy array or Pandas"
                    + " series, returned {} instead".format(len(mask.shape))
                )
        if np.sum(mask) == 0:
            warn("No samples in subset, returning NaN", RuntimeWarning)
            return np.nan
        X = X[mask]
        y_pred = estimator.predict(X)
        return score(y_true[mask], y_pred, **kwargs)

    return sliced_metric



__all__ = ["GMMClassifier", "BayesianGMMClassifier", "GMMOutlierDetector", "BayesianGMMOutlierDetector"]

from sklego.mixture.bayesian_gmm_classifier import BayesianGMMClassifier
from sklego.mixture.bayesian_gmm_detector import BayesianGMMOutlierDetector
from sklego.mixture.gmm_classifier import GMMClassifier
from sklego.mixture.gmm_outlier_detector import GMMOutlierDetector



import numbers
from datetime import timedelta
from itertools import combinations
from warnings import warn

import narwhals.stable.v1 as nw
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.model_selection._split import _BaseKFold
from sklearn.utils.validation import indexable
from sklearn_compat.utils.validation import check_array

from sklego.base import Clusterer
from sklego.common import sliding_window


class TimeGapSplit:
    r"""Provides train/test indices to split time series data samples.

    This cross-validation object is a variation of TimeSeriesSplit with the following  differences:

    - The splits are made based on datetime duration, instead of number of rows.
    - The user specifies the `valid_duration` and either `train_duration` or `n_splits`.
    - The user can specify a `gap_duration` that is added after the training split and before the validation split.

    The 3 duration parameters can be used to really replicate how the model is going to be used in production in batch
    learning.

    Each validation fold doesn't overlap. The entire `window` moves by 1 `valid_duration` until there is not enough
    data.

    If this would lead to more splits than specified with `n_splits`, the `window` moves by `valid_duration` times the
    fraction of possible splits and requested splits:

    - `n_possible_splits = (total_length - train_duration - gap_duration) // valid_duration`
    - `time_shift = valid_duration * n_possible_splits / n_splits`

    so the CV spans the whole dataset.

    If `train_duration` is not passed but `n_splits` is, the training duration is increased to:

    `train_duration = total_length - (gap_duration + valid_duration * n_splits)`

    such that the shifting the entire window by one validation duration spans the whole training set.

    Parameters
    ----------
    date_serie : Series
        Series with the date, that should have all the indices of X used in the split() method.
        If the Series is not pandas-like (for example, if it's a Polars Series, which does not have
        an index) then it must the same length as the `X` and `y` objects passed to `split`.
    valid_duration : datetime.timedelta
        Retraining period.
    stride_duration : datetime.timedelta | None
        The time shift for the training period. If `None`, then it fallbacks to `valid_duration` value.
    train_duration : datetime.timedelta | None, default=None
        Historical training data.
    gap_duration : datetime.timedelta, default=timedelta(0)
        Forward-looking window of the target. The period of the forward-looking window necessary to create your target
        variable. This period is dropped at the end of your training folds due to lack of recent data. In production you
        would have not been able to create the target for that period, and you would have to drop it from the training data.
    n_splits : int | None, default=None
        Number of splits.
    window : Literal["rolling", "expanding"], default="rolling"
        Type of moving window to use.

        - `"rolling"` window has fixed size and is shifted entirely.
        - `"expanding"` left side of window is fixed, right border increases each fold.

    Notes
    -----
    Native cross-dataframe support is achieved using
    [Narwhals](https://narwhals-dev.github.io/narwhals/){:target="_blank"}.
    Supported dataframes are:

    - pandas
    - Polars (eager)
    - Modin
    - cuDF

    See [Narwhals docs](https://narwhals-dev.github.io/narwhals/extending/){:target="_blank"} for an up-to-date list
    (and to learn how you can add your dataframe library to it!), though note that only those
    convertible to `numpy` arrays will work with this class.
    """

    def __init__(
        self,
        date_serie,
        valid_duration,
        stride_duration=None,
        train_duration=None,
        gap_duration=timedelta(0),
        n_splits=None,
        window="rolling",
    ):
        # If stride length is not defined, set it equal to the length validation set
        if stride_duration is None:
            stride_duration = valid_duration

        if stride_duration <= timedelta(milliseconds=0):
            msg = f"`stride_duration` should be a positive timedelta, found {stride_duration}"
            raise ValueError(msg)

        if (train_duration is None) and (n_splits is None):
            raise ValueError("Either train_duration or n_splits have to be defined")

        if (train_duration is not None) and (train_duration <= gap_duration):
            raise ValueError("gap_duration is longer than train_duration, it should be shorter.")

        if (train_duration is not None) and (train_duration <= stride_duration):
            raise ValueError("stride_duration is longer than train_duration, it should be shorter.")

        self.date_serie = nw.from_native(date_serie, series_only=True).alias("__date__")
        self.train_duration = train_duration
        self.valid_duration = valid_duration
        self.gap_duration = gap_duration
        self.stride_duration = stride_duration
        self.n_splits = n_splits
        self.window = window

    def _join_date_and_x(self, X):
        """Creates a DataFrame indexed by the pandas index (the same as `date_serie`) with date column joined with that
        index and with the 'numpy index' column (i.e. just a range) that is required for the output and the rest of
        sklearn.

        If the user is working with index-less dataframes (e.g. Polars), then `self.date_series` needs to be the same
        length as `X`.

        Parameters
        ----------
        X : DataFrame
            Dataframe with the data to split
        """
        X_index_df = nw.maybe_align_index(self.date_serie, X).to_frame().with_row_index("np_index")

        return X_index_df

    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : DataFrame
            Dataframe with the data to split.
        y : array-like | None, default=None
            Ignored, present for compatibility.
        groups : array-like | None, default=None
            Ignored, present for compatibility.

        Yields
        -------
        tuple[np.ndarray, np.ndarray]
            Train and test indices of the same fold.
        """

        X = nw.from_native(X, eager_only=True)
        X_index_df = self._join_date_and_x(X)
        X_index_df = X_index_df.sort("__date__", descending=False)

        if len(X) != len(X_index_df):
            raise AssertionError(
                "X and X_index_df are not the same length, there must be some index missing in 'self.date_serie'"
            )

        date_min = X_index_df["__date__"].min()
        date_max = X_index_df["__date__"].max()
        date_length = X_index_df["__date__"].max() - X_index_df["__date__"].min()

        if (self.train_duration is None) and (self.n_splits is not None):
            self.train_duration = date_length - (
                self.gap_duration + self.valid_duration + (self.n_splits - 1) * self.stride_duration
            )

        if (self.train_duration is not None) and (self.train_duration <= self.gap_duration):
            raise ValueError("gap_duration is longer than train_duration, it should be shorter.")

        n_split_max = (
            1 + (date_length - self.train_duration - self.gap_duration - self.valid_duration) / self.stride_duration
        )

        if self.n_splits:
            if n_split_max < self.n_splits:
                raise ValueError(
                    (
                        "Number of folds requested = {1} are greater"
                        " than maximum  ={0} possible"
                        " based on the given values."
                    ).format(n_split_max, self.n_splits)
                )

        current_date = date_min
        start_date = date_min
        # if the n_splits is smaller than what would usually be done for train, validation, stride and gap duration,
        # the next fold is slightly further in time than just stride_duration
        if self.n_splits is not None:
            time_shift = self.stride_duration * n_split_max / self.n_splits
        else:
            time_shift = self.stride_duration
        while True:
            if current_date + self.train_duration + time_shift + self.gap_duration > date_max:
                break

            X_train_df = X_index_df.filter(
                nw.col("__date__") >= start_date, nw.col("__date__") < current_date + self.train_duration
            )
            X_valid_df = X_index_df.filter(
                nw.col("__date__") >= current_date + self.train_duration + self.gap_duration,
                nw.col("__date__") < current_date + self.train_duration + self.valid_duration + self.gap_duration,
            )

            current_date = current_date + time_shift
            if self.window == "rolling":
                start_date = current_date
            yield (
                X_train_df["np_index"].to_numpy(),
                X_valid_df["np_index"].to_numpy(),
            )

    def get_n_splits(self, X=None, y=None, groups=None):
        """Get the number of splits

        Parameters
        ----------
        X : DataFrame
            Dataframe with the data to split.
        y : array-like | None, default=None
            Ignored, present for compatibility.
        groups : array-like | None, default=None
            Ignored, present for compatibility.

        Returns
        -------
        int
            Number of splits.
        """
        return sum(1 for x in self.split(X, y, groups))

    def summary(self, X):
        """Describe all folds.

        Parameters
        ----------
        X : DataFrame
            Dataframe with the data to split.

        Returns
        -------
        DataFrame
            Summary of all folds.
        """
        summary = []
        X = nw.from_native(X, eager_only=True)
        X_index_df = self._join_date_and_x(X)

        summary = {
            "Start date": [],
            "End date": [],
            "Period": [],
            "Unique days": [],
            "nbr samples": [],
            "part": [],
            "fold": [],
        }
        native_namespace = nw.get_native_namespace(X)

        def update_split_info(indices, j, part, summary):
            dates = X_index_df["__date__"][indices]
            mindate = dates.min()
            maxdate = dates.max()
            n_unique = dates.n_unique()

            summary["Start date"].append(mindate)
            summary["End date"].append(maxdate)
            summary["Period"].append(maxdate - mindate)
            summary["Unique days"].append(n_unique)
            summary["nbr samples"].append(len(indices))
            summary["part"].append(part)
            summary["fold"].append(j)

        j = 0
        for i in self.split(nw.to_native(X)):
            train_info = nw.to_native(nw.new_series(name="tmp", values=i[0], native_namespace=native_namespace))
            valid_info = nw.to_native(nw.new_series(name="tmp", values=i[1], native_namespace=native_namespace))
            update_split_info(train_info, j, "train", summary)
            update_split_info(valid_info, j, "valid", summary)
            j = j + 1

        result = nw.from_dict(summary, native_namespace=native_namespace)
        result = nw.maybe_set_index(result, ["fold", "part"])
        return nw.to_native(result)


def KlusterFoldValidation(**kwargs):
    warn(
        "Please use `ClusterFoldValidation` instead of `KlusterFoldValidation`."
        "We will use correct spelling going forward and `KlusterFoldValidation` will be deprecated.",
        DeprecationWarning,
    )
    return ClusterFoldValidation(**kwargs)


class ClusterFoldValidation:
    """Cross validator that creates folds based on provided cluster method.
    This ensures that data points in the same cluster are not split across different folds.

    !!! info "New in version 0.8.2"

    Parameters
    ----------
    cluster_method : Clusterer
        Clustering method to use for the fold validation.
    """

    def __init__(self, cluster_method=None):
        if not isinstance(cluster_method, Clusterer):
            raise ValueError("The KlusterFoldValidation only works on cluster methods with .fit_predict.")

        self.cluster_method = cluster_method
        self.n_splits = None

    def split(self, X, y=None, groups=None):
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Array to split on.
        y : array-like of shape (n_samples,) | None, default=None
            Ignored, present for compatibility.
        groups : array-like of shape (n_samples,) | None, default=None
            Ignored, present for compatibility.

        Yields
        -------
        tuple[np.ndarray, np.ndarray]
            Train and test indices of the same fold.
        """

        X = check_array(X)

        if not self._method_is_fitted(X):
            self.cluster_method.fit(X)
        clusters = self.cluster_method.predict(X)

        self.n_splits = len(np.unique(clusters))

        if self.n_splits < 2:
            raise ValueError(f"Clustering method resulted in {self.n_splits} cluster, too few for fold validation")

        for label in np.unique(clusters):
            yield (
                np.where(clusters != label)[0],
                np.where(clusters == label)[0],
            )

    def _method_is_fitted(self, X):
        """Check if the `cluster_method` is fitted

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Array to use if the method is fitted

        Returns
        -------
        bool
            True if fitted, else False
        """

        try:
            self.cluster_method.predict(X[0:1, :])
            return True
        except NotFittedError:
            return False


class GroupTimeSeriesSplit(_BaseKFold):
    """Sliding window time series split.

    Create `n_splits` folds with an as equally possible size through a smart variant of a brute force search.
    Groups parameter in `.split()` method should be filled with the time groups (e.g. years)

    If `n_splits` is 3 ("*" = train, "x" = test):

    ```console
    * * * x x x - - - - -
    - - - * * * x x x - -
    - - - - - - * * * x x
    ```

    Parameters
    ----------
    n_splits : int
        Amount of (train, test) splits to generate.
    """

    # table above inspired by sktime

    def __init__(self, n_splits):
        if not isinstance(n_splits, numbers.Integral):
            raise ValueError(
                "The number of folds must be of Integral type. %s of type %s was passed." % (n_splits, type(n_splits))
            )
        n_splits = int(n_splits)

        if n_splits <= 1:
            raise ValueError(
                "k-fold cross-validation requires at least one"
                " train/test split by setting n_splits=2 or more,"
                " got n_splits={0}.".format(n_splits)
            )

        self.n_splits = n_splits

    def summary(self):
        """Describe all folds in a pd.DataFrame which displays the groups splits and extra statistics about it.

        Can only be run after having applied the `.split()` method to the `GroupTimeSeriesSplit` instance.

        Returns
        -------
        pd.DataFrame
            Summary of all folds.
        """
        try:
            return (
                self._grouped_df.sort_index()
                .assign(group=lambda df: df["group"].astype(int))
                .assign(obs_per_group=lambda df: df.groupby("group")["observations"].transform("sum"))
                .assign(ideal_group_size=round(self._ideal_group_size))
                .assign(diff_from_ideal_group_size=lambda df: df["obs_per_group"] - df["ideal_group_size"])
            )
        except AttributeError:
            raise AttributeError(".summary() only works after having ran .split(X, y, groups).")

    def split(self, X=None, y=None, groups=None):
        """Generate the train-test splits of all the folds

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features), default=None
            Data to split.
        y : array-like of shape (n_samples,), default=None
            The target variable for supervised learning problems.
        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into train/test set,

        Returns
        -------
        List[np.ndarray]
            List containing train-test split indices of each fold.
        """
        if groups is None:
            raise ValueError("Groups cannot be None")
        X, y, groups = indexable(X, y, groups)
        n_groups = np.unique(groups).shape[0]
        if self.n_splits >= n_groups:
            raise ValueError(
                ("n_splits({0}) must be less than the amount of unique groups({1}).").format(self.n_splits, n_groups)
            )
        return list(self._iter_test_indices(X, y, groups))

    def get_n_splits(self, X=None, y=None, groups=None):
        """Get the amount of splits

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features), default=None
            Ignored, present for compatibility.
        y : array-like of shape (n_samples,), default=None
            Ignored, present for compatibility.
        groups : array-like of shape (n_samples,), default=None
            Ignored, present for compatibility.

        Returns
        -------
        int
            Amount of splits.
        """
        return self.n_splits

    def _check_for_long_estimated_runtime(self, groups):
        """Check for combinations of n_splits and unique groups and raises `UserWarning` if runtime is expected to take
        over one minute.

        Parameters
        ----------
        groups : array-like of shape (n_samples,)
            Groups to check for unique groups and n_splits.

        Raises
        ------
        UserWarning
            If runtime is expected to take over one minute.
        """
        unique_groups = len(set(groups))
        warning = (
            "Finding the optimal split points with {0} unique groups and n_splits at {1} can take several minutes."
        ).format(unique_groups, self.n_splits)
        if self.n_splits == 4 and unique_groups > 250:
            warn(
                warning + " Consider to decrease n_splits to 3 or lower.",
                UserWarning,
            )

        elif self.n_splits == 5 and unique_groups > 100:
            warn(
                warning + " Consider to decrease n_splits to 4 or lower.",
                UserWarning,
            )

        elif self.n_splits > 5 and unique_groups > 30:
            warn(
                warning + " Consider to decrease n_splits to 5 or lower.",
                UserWarning,
            )

    def _iter_test_indices(self, X=None, y=None, groups=None):
        """Calculate the optimal division of groups into folds so that every fold is as equally large as possible.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features), default=None
            Ignored, present for compatibility.
        y : array-like of shape (n_samples,), default=None
            Ignored, present for compatibility.
        groups : array-like of shape (n_samples,), default=None
            Array with groups.

        Yields
        ------
        tuple[np.ndarray, np.ndarray]
            Train and test indices of the same fold.
        """
        self._check_for_long_estimated_runtime(groups)
        self._first_split_index, self._last_split_index = self._calc_first_and_last_split_index(groups=groups)
        self._best_splits = self._get_split_indices()
        groups = self._regroup(groups)
        for i in range(self.n_splits):
            yield np.where(groups == i)[0], np.where(groups == i + 1)[0]

    def _calc_first_and_last_split_index(self, X=None, y=None, groups=None):
        """Calculate an approximate first and last split point to reduce the amount of options during a brute force
        search.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features), default=None
            Ignored, present for compatibility.
        y : array-like of shape (n_samples,), default=None
            Ignored, present for compatibility.
        groups : array-like of shape (n_samples,), default=None
            Array with groups.

        Returns
        -------
        tuple[int, int]
            Approximate first and last split indexes.
        """

        # get the counts (=amount of rows) for each group
        self._grouped_df = (
            pd.DataFrame(np.array(groups))
            .rename(columns={0: "index"})
            .groupby("index")
            .size()
            .sort_index()
            .to_frame()
            .rename(columns={0: "observations"})
        )

        # set the ideal group_size and reduce it to 90% to have some leverage
        self._ideal_group_size = np.sum(self._grouped_df["observations"]) / (self.n_splits + 1)
        init_ideal_group_size = self._ideal_group_size * 0.9

        # initialize the index of the first split, to reduce the amount of possible index split options
        first_split_index = (
            self._grouped_df.assign(cumsum_obs=lambda df: df["observations"].cumsum())
            .assign(group_id=lambda df: (df["cumsum_obs"] - 1) // init_ideal_group_size)
            .reset_index()
            .loc[lambda df: df["group_id"] != 0]
            .iloc[0]
            .name
        )
        # initialize the index of the last split point, to reduce the amount of possible index split options
        last_split_index = len(self._grouped_df) - (
            self._grouped_df.assign(
                observations=lambda df: df["observations"].to_numpy()[::-1],
                cumsum_obs=lambda df: df["observations"].cumsum(),
            )
            .reset_index()
            .assign(group_id=lambda df: (df["cumsum_obs"] - 1) // init_ideal_group_size)
            .loc[lambda df: df["group_id"] != 0]
            .iloc[0]
            .name
            - 1
        )
        return first_split_index, last_split_index

    def _get_split_indices(self):
        """Calculate for each possible splits the total absolute different of the groups to the ideal group size and
        saves the split with the least absolute difference.

        Returns
        -------
        int
            Index of the best split point.
        """
        # set the index range to search possible splits for
        index_range = range(self._first_split_index, self._last_split_index)

        observations = self._grouped_df["observations"].tolist()

        # create generator with all the possible index splits
        # e.g. for [0, 1, 3, 5, 8] and self.n_splits = 2
        # [(1,2), (1,3), (1,4), (2,3), (2,4), (3,4)]
        # with for the first split:
        # group1 = [:1]
        # group2 = [1:2]
        # group3 = [2:]
        splits_generator_shifted = combinations(index_range, self.n_splits)

        # get the first iteration
        first_splits = next(splits_generator_shifted)

        # create a new generator that starts from the beginning again
        splits_generator = combinations(index_range, self.n_splits)

        # generate a list, with for every group the difference between them and the ideal group size, e.g.
        # ideal_group_size = 100
        # group_sizes = [10,20,270]
        # diff_from_ideal_list = [-90, -80, 170]
        diff_from_ideal_list = [sum(observations[: first_splits[0]]) - self._ideal_group_size]
        for split in sliding_window(first_splits, window_size=2, step_size=1):
            try:
                diff_from_ideal_list += [sum(observations[split[0] : split[1]]) - self._ideal_group_size]
            except IndexError:
                diff_from_ideal_list += [sum(observations[split[0] :]) - self._ideal_group_size]

        # keep track of the minimum of the total difference from all groups to the ideal group size
        min_diff = sum([abs(diff) for diff in diff_from_ideal_list])
        best_splits = first_splits

        # loop through all possible split points and check whether a new split
        # has a less total difference from all groups to the ideal group size
        for prev_splits, new_splits in zip(splits_generator, splits_generator_shifted):
            diff_from_ideal_list = self._calc_new_diffs(observations, diff_from_ideal_list, prev_splits, new_splits)
            new_diff = sum([abs(diff) for diff in diff_from_ideal_list])

            # if with the new split the difference is less than the current most optimal, save the new split
            if new_diff < min_diff:
                min_diff = new_diff
                best_splits = new_splits
        return best_splits

    @staticmethod
    def _calc_new_diffs(values, diff_list, prev_splits, new_splits):
        """Calculate the new group size differences compared to the optimal group size.

        Parameters
        ----------
        values : list
            List of values.
        diff_list : list
            List of values with for each index split its difference from the optimal group size.
        prev_splits : tuple
            Indices of the previous splits, excluding index 0 and the last index.
        new_splits : tuple
            Indices of the new splits, excluding index 0 and the last index.

        Returns
        -------
        list
            List of values with for each index split its difference from the optimal group size.
        """
        # calculate which indices have changed, e.g.:
        # new_index = (1,2,5)
        # prev_index = (1,2,4)
        # index_diffs = (0,0,1)
        index_diffs = [new_index - prev_index for prev_index, new_index in zip(prev_splits, new_splits)]
        new_diff_list = diff_list.copy()

        # calculate the effects of the index change to the groups
        for index, diff in enumerate(index_diffs):
            if diff != 0:
                start_index, end_index = (
                    (prev_splits[index], new_splits[index])
                    if prev_splits[index] < new_splits[index]
                    else (new_splits[index], prev_splits[index])
                )

                # calculate the value change from one group to another
                value_change = sum(values[start_index:end_index])

                # if diff < 0 the previous group gains values, so change value_change to -value_change
                value_change = value_change if diff > 0 else -value_change

                # change the values of the current and next group
                new_diff_list[index] += value_change
                new_diff_list[index + 1] -= value_change

        return new_diff_list

    def _regroup(self, groups):
        """Specify in which group every observation belongs.

        Parameters
        ----------
        groups : array-like of shape (n_samples,)
            Array with original groups.

        Returns
        -------
        np.ndarray
            Indices for the train and test splits of each fold
        """

        df = self._grouped_df.copy().reset_index()
        # set each unique group to the right group_id to group them into folds
        df.loc[: self._best_splits[0], "group"] = 0
        for group_id, splits in enumerate(sliding_window(self._best_splits, 2, 1)):
            try:
                df.loc[splits[0] : splits[1], "group"] = group_id + 1
            except IndexError:
                df.loc[splits[0] :, "group"] = group_id + 1

        self._grouped_df = df
        # create a mapper to set every group to the right group_id
        mapper = dict(zip(df["index"], df["group"]))
        return np.vectorize(mapper.get)(groups)



import warnings

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.feature_selection import f_classif, f_regression
from sklearn.feature_selection._base import SelectorMixin
from sklearn.utils.validation import check_is_fitted
from sklearn_compat.utils.validation import validate_data


def _redundancy_pearson(X, selected, left):
    """Redundancy function for the MRMR feature selector algorithm

    Parameters
    ----------
    X : array-like, shape=(n_samples, n_features,)
        Training data. Used to compute redundancy of the training features.
    selected : array-like.
        List of indexes of the selected features at iteration i-th.
    left : array-like.
        List of indexes of the left features at iteration i-th. Mrmr will select a feature
        from this list.

    Returns
    -------
    np.ndarray, shape = (len(left), )
        The array containing the redundancy score using pearson correlation.
    """
    # if len(selected) == 0:
    #     return np.ones(len(left))

    X_norm = X - np.mean(X, axis=0, keepdims=True)
    Xs = X_norm[:, selected]
    Xl = X_norm[:, left]

    num = (Xl[:, None, :] * Xs[:, :, None]).sum(axis=0)
    den = np.sqrt((Xl[:, None, :] ** 2).sum(axis=0)) * np.sqrt((Xs[:, :, None] ** 2).sum(axis=0))

    return np.sum(np.abs(np.nan_to_num(num / den, nan=np.finfo(float).eps)), axis=0)


class MaximumRelevanceMinimumRedundancy(SelectorMixin, BaseEstimator):
    r"""Maximum Relevance Minimum Redundancy (MRMR) is an iterative feature selection method commonly used in data
    science to select a subset of features from a larger feature set. The goal of MRMR is to choose features that
    have high relevance to the target variable while minimizing redundancy among the already selected features.

    How MRMR works:

    1. Compute the relevance of each feature to the target variable: The relevance of a feature is typically
    measured using a metric such as mutual information, correlation coefficient, or another appropriate measure of
    dependence between the feature and the target variable.

    2. Compute the redundancy between each pair of features: Redundancy is the degree of similarity or overlap between
    features. It can be measured using metrics such as mutual information, correlation coefficient, or other similarity
    measures.

    3. Select features based on the maximum relevance and minimum redundancy criteria: MRMR aims to maximize the
    relevance of selected features to the target variable while minimizing redundancy among them.

    4. Construct the final subset of features: MRMR iteratively adds features to the selected subset until a predefined
    number of features is reached.

    The implemented formula is:

    $$\text{score}_{i}(f) = \frac{\text{relevance}(f | y)}{\text{redundancy}(f | \text{selected}_{i-1})}$$

    !!! warning
        If a custom relevance_func is provided it must have this signature:
        `Callable[[np.ndarray, np.ndarray], np.ndarray]`
        It should accept X, y as arguments and it should compute the score for each feature of X
        and return an array of shape (n_features_in_,).

    !!! warning
        If a custom redundancy_func is provided it must have the same signature as the method _redundancy_pearson, hence
        the function must have three parameters:

            - X : array-like, shape=(n_samples, n_features,). Training used to compute redundancy of the training features.

            - selected : array-like. List of indexes of the selected features at iteration i-th.

            - left : array-like. List of indexes of the left features at iteration i-th. Mrmr will select a feature from this list.

        and it must return:

            - np.ndarray, shape = (len(left), ), The array containing the redundancy score using the custom function.

    !!! info "New in version 0.8.0"

    Parameters
    ----------
    k : int
        Number of feature the model should use.
    relevance_func : str | Callable, default="f"
        The relevance function to use. The default maps to scikit-learn [f_classif](https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.f_classif.html) or  [f_regression](https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.f_regression.html) for classification or regression (resp.)
    redundancy_func : str | Callable, default="p"
        The redundancy function to use. The default maps to Pearson correlation computed for each remaining features.
    kind : Literal["auto", "classficiation", "regression"], default="auto".
        'classification' or 'regression' or 'auto' if auto the model
        will try to infer the type of problem looking at the y data type, by default "auto".

    Attributes
    ----------
    _y_dtype : np.dtype
        data type of y
    selected_features_ : array-like of shape (k,)
        Indexes of the selected features.
    scores_ : array-like of shape (k,)
        Scores of the selected features.

    Examples
    --------
    ```py
    from sklego.feature_selection import MaximumRelevanceMinimumRedundancy
    from sklearn.datasets import make_classification

    mrmr =  MaximumRelevanceMinimumRedundancy(k=4,
            kind='auto',
            redundancy_func='p',
            relevance_func='f')

    X, y = make_classification(n_features=4)

    # Fit mrmr model
    mrmr = mrmr.fit(X, y)

    # Selected features
    selected_features = mrmr.selected_features_

    # Get the scores of the selected features
    feature_scores = mrmr.scores_
    ```
    """

    _required_parameters = ["k"]

    def __init__(self, k, *, relevance_func="f", redundancy_func="p", kind="auto"):
        self.k = k
        self.relevance_func = relevance_func
        self.redundancy_func = redundancy_func
        self.kind = kind

    def _get_support_mask(self):
        """SelectorMixin base function to get the selected features mask

        Returns
        -------
        np.ndarray
            Array of boolean, mask indicating if feature n is selected by mrmr or not.
        """
        check_is_fitted(self, ["selected_features_"])
        all_features = np.arange(0, self.n_features_in_)
        return np.isin(all_features, self.selected_features_)

    @property
    def _get_relevance(self):
        """get relevance function from init values."""
        if self.relevance_func == "f":
            if (self.kind == "auto" and np.issubdtype(self._y_dtype, np.integer)) | (self.kind == "classification"):
                return lambda X, y: np.nan_to_num(f_classif(X, y)[0])
            elif (self.kind == "auto" and np.issubdtype(self._y_dtype, np.floating)) | (self.kind == "regression"):
                return lambda X, y: np.nan_to_num(f_regression(X, y)[0])
            else:
                raise ValueError(
                    "`kind` parameter must be 'auto', 'classification' or 'regression' and y dtype must be numeric"
                )
        elif callable(self.relevance_func):
            return self.relevance_func
        else:
            raise ValueError(f"Relevance function supported are 'f' or Callable, got {self.relevance_func}")

    @property
    def _get_redundancy(self):
        """get redundancy function from init values."""
        if self.redundancy_func == "p":
            return _redundancy_pearson
        elif callable(self.redundancy_func):
            return self.redundancy_func
        else:
            raise ValueError(f"Redundancy function supported are 'p' or Callable, got {self.redundancy_func}")

    def fit(self, X, y):
        """Fit the underlying feature selection algorithm on the training data `X` and `y`
        using the provided redundancy and relevance functions.


        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        self : MaximumRelevanceMinimumRedundancy
            The fitted estimator.

        Raises
        ------
        ValueError
            if:

                k parameter is not integer type or is < n_features_in (X.shape[1]) or < 1
        """
        X, y = validate_data(self, X=X, y=y, dtype="numeric", y_numeric=True, reset=True)
        self._y_dtype = y.dtype

        relevance = self._get_relevance
        redundancy = self._get_redundancy

        left_features = list(range(self.n_features_in_))
        selected_features = []
        selected_scores = []

        if not isinstance(self.k, int):
            raise ValueError("k parameter must be integer type")
        if self.k > self.n_features_in_:
            raise ValueError(f"k ({self.k}) parameter must be less than n_features_in_ ({self.n_features_in_})")
        elif self.k == self.n_features_in_:
            warnings.warn("k parameter is equal to n_features_in, no feature selection is applied")
            return np.asarray(left_features)
        elif self.k < 1:
            raise ValueError(f"k ({self.k}) parameter must be greater than or equal to 1")

        # computed one time for all features

        rel_score = relevance(X, y)

        for i in range(self.k):
            red_i = redundancy(X, selected_features, left_features) / i if i > 0 else 1
            mrmr_score_i = rel_score[left_features] / red_i
            selected_index = np.argmax(mrmr_score_i)
            selected_features += [left_features.pop(selected_index)]
            selected_scores += [mrmr_score_i[selected_index]]
        self.selected_features_ = np.asarray(selected_features)
        self.scores_ = np.asarray(selected_scores)
        return self



import pytest

from sklego.datasets import load_abalone, load_chicken, load_hearts, load_penguins, make_simpleseries


def test_chickweight1():
    X, y = load_chicken(return_X_y=True)
    assert X.shape == (578, 3)
    assert y.shape[0] == 578


def test_chickweight2():
    df = load_chicken(as_frame=True)
    assert df.shape == (578, 4)


def test_abalone1():
    X, y = load_abalone(return_X_y=True)
    assert X.shape == (4177, 8)
    assert y.shape[0] == 4177


def test_abalone2():
    df = load_abalone(as_frame=True)
    assert df.shape == (4177, 9)


def test_simpleseries_constant_season():
    df = make_simpleseries(
        n_samples=365 * 2,
        as_frame=True,
        start_date="2018-01-01",
        trend=0,
        noise=0,
        season_trend=0,
    ).assign(month=lambda d: d["date"].dt.month, year=lambda d: d["date"].dt.year)
    agg = df.groupby(["year", "month"]).mean().reset_index()
    var = agg.loc[lambda d: d["month"] == 1]["yt"].var()
    assert var == pytest.approx(0.0, abs=0.01)


def test_load_hearts():
    df = load_hearts(as_frame=True)
    assert df.shape == (303, 14)


def test_penguin1():
    X, y = load_penguins(return_X_y=True)
    assert X.shape == (344, 6)
    assert y.shape[0] == 344


def test_penguin2():
    df = load_penguins(as_frame=True)
    assert df.shape == (344, 7)



__all__ = [
    "ColumnCapper",
    "ColumnDropper",
    "ColumnSelector",
    "DictMapper",
    "FormulaicTransformer",
    "IdentityTransformer",
    "InformationFilter",
    "IntervalEncoder",
    "LinearEmbedder",
    "OrthogonalTransformer",
    "OutlierRemover",
    "PandasTypeSelector",
    "TypeSelector",
    "RandomAdder",
    "RepeatingBasisFunction",
    "MonotonicSplineTransformer",
]

from sklego.preprocessing.columncapper import ColumnCapper
from sklego.preprocessing.dictmapper import DictMapper
from sklego.preprocessing.formulaictransformer import FormulaicTransformer
from sklego.preprocessing.identitytransformer import IdentityTransformer
from sklego.preprocessing.intervalencoder import IntervalEncoder
from sklego.preprocessing.linearembedder import LinearEmbedder
from sklego.preprocessing.monotonicspline import MonotonicSplineTransformer
from sklego.preprocessing.outlier_remover import OutlierRemover
from sklego.preprocessing.pandastransformers import ColumnDropper, ColumnSelector, PandasTypeSelector, TypeSelector
from sklego.preprocessing.projections import InformationFilter, OrthogonalTransformer
from sklego.preprocessing.randomadder import RandomAdder
from sklego.preprocessing.repeatingbasis import RepeatingBasisFunction



import logging
from inspect import signature

import numpy as np
from sklearn.base import BaseEstimator, MetaEstimatorMixin, RegressorMixin, clone, is_classifier, is_regressor
from sklearn.exceptions import NotFittedError
from sklearn.utils.metaestimators import available_if
from sklearn.utils.validation import _check_sample_weight, check_is_fitted
from sklearn_compat.utils.validation import validate_data


class ZeroInflatedRegressor(RegressorMixin, MetaEstimatorMixin, BaseEstimator):
    """A meta regressor for zero-inflated datasets, i.e. the targets contain a lot of zeroes.

    `ZeroInflatedRegressor` consists of a classifier and a regressor.

    - The classifier's task is to find if the target is zero or not.
    - The regressor's task is to output a (usually positive) prediction whenever the classifier indicates that
    there should be a non-zero prediction.

    The regressor is only trained on examples where the target is non-zero, which makes it easier for it to focus.

    At prediction time, the classifier is first asked if the output should be zero. If yes, output zero.
    Otherwise, ask the regressor for its prediction and output it.

    Parameters
    ----------
    classifier : scikit-learn compatible classifier
        A classifier that answers the question "Should the output be zero?".
    regressor : scikit-learn compatible regressor
        A regressor for predicting the target. Its prediction is only used if `classifier` says that the output is
        non-zero.
    handle_zero : Literal["error", "ignore"], default="error"
            How to behave in the case that all train set output consists of zero values only.

            - `handle_zero = 'error'`: will raise a `ValueError` (default).
            - `handle_zero = 'ignore'`: will continue to train the regressor on the entire dataset.

    Attributes
    ----------
    classifier_ : scikit-learn compatible classifier
        The fitted classifier.
    regressor_ : scikit-learn compatible regressor
        The fitted regressor.

    Examples
    --------
    ```py
    import numpy as np
    from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
    from sklego.meta import ZeroInflatedRegressor

    np.random.seed(0)
    X = np.random.randn(10000, 4)
    y = ((X[:, 0]>0) & (X[:, 1]>0)) * np.abs(X[:, 2] * X[:, 3]**2)

    model = ZeroInflatedRegressor(
        classifier=ExtraTreesClassifier(random_state=0, max_depth=10),
        regressor=ExtraTreesRegressor(random_state=0)
    ).fit(X, y)

    model.predict(X[:5])
    # array([4.91483294, 0.        , 0.        , 0.04941909, 0.        ])

    model.score_samples(X[:5]).round(2)
    # array([3.73, 0.  , 0.11, 0.03, 0.06])
    ```
    """

    _required_parameters = ["classifier", "regressor"]

    def __init__(self, classifier, regressor, handle_zero="error") -> None:
        self.classifier = classifier
        self.regressor = regressor
        self.handle_zero = handle_zero

    def fit(self, X, y, sample_weight=None):
        """Fit the underlying classifier and regressor using `X` and `y` as training data. The regressor is only trained
        on examples where the target is non-zero.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features )
            The training data.
        y : array-like of shape (n_samples,)
            The target values.
        sample_weight : array-like of shape (n_samples, ), default=None
            Individual weights for each sample.

        Returns
        -------
        self : ZeroInflatedRegressor
            The fitted estimator.

        Raises
        ------
        ValueError
            If `classifier` is not a classifier
            If `regressor` is not a regressor
            If all train target entirely consists of zeros and `handle_zero="error"`
        """
        X, y = validate_data(self, X=X, y=y, reset=True)

        if not is_classifier(self.classifier):
            raise ValueError(
                f"`classifier` has to be a classifier. Received instance of {type(self.classifier)} instead."
            )
        if not is_regressor(self.regressor):
            raise ValueError(f"`regressor` has to be a regressor. Received instance of {type(self.regressor)} instead.")
        if self.handle_zero not in {"ignore", "error"}:
            raise ValueError(
                f"`handle_zero` has to be one of {'ignore', 'error'}. Received '{self.handle_zero}' instead."
            )

        sample_weight = _check_sample_weight(sample_weight, X)
        try:
            check_is_fitted(self.classifier)
            self.classifier_ = self.classifier
        except NotFittedError:
            self.classifier_ = clone(self.classifier)

            if "sample_weight" in signature(self.classifier_.fit).parameters:
                self.classifier_.fit(X, y != 0, sample_weight=sample_weight)
            else:
                logging.warning("Classifier ignores sample_weight.")
                self.classifier_.fit(X, y != 0)

        indices_for_training = np.where(y != 0)[0]  # these are the non-zero indices
        if (self.handle_zero == "ignore") & (
            indices_for_training.size == 0
        ):  # if we choose to ignore that all train set output is 0
            logging.warning("Regressor will be training on `y` consisting of zero values only.")
            indices_for_training = np.where(y == 0)[0]  # use the whole train set

        if indices_for_training.size > 0:
            try:
                check_is_fitted(self.regressor)
                self.regressor_ = self.regressor
            except NotFittedError:
                self.regressor_ = clone(self.regressor)

                if "sample_weight" in signature(self.regressor_.fit).parameters:
                    self.regressor_.fit(
                        X[indices_for_training],
                        y[indices_for_training],
                        sample_weight=sample_weight[indices_for_training] if sample_weight is not None else None,
                    )
                else:
                    logging.warning("Regressor ignores sample_weight.")
                    self.regressor_.fit(
                        X[indices_for_training],
                        y[indices_for_training],
                    )
        else:
            raise ValueError(
                """The predicted training labels are all zero, making the regressor obsolete. Change the classifier
                or use a plain regressor instead. Alternatively, you can choose to ignore that predicted labels are
                all zero by setting flag handle_zero = 'ignore'"""
            )

        return self

    def predict(self, X):
        """Predict target values for `X` using fitted estimator by first asking the classifier if the output should be
        zero. If yes, output zero. Otherwise, ask the regressor for its prediction and output it.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples,)
            The predicted values.
        """
        check_is_fitted(self, ["n_features_in_", "classifier_", "regressor_"])
        X = validate_data(self, X=X, reset=False)

        output = np.zeros(len(X))
        non_zero_indices = np.where(self.classifier_.predict(X))[0]

        if non_zero_indices.size > 0:
            output[non_zero_indices] = self.regressor_.predict(X[non_zero_indices])

        return output

    @available_if(lambda self: hasattr(self.classifier_, "predict_proba"))
    def score_samples(self, X):
        r"""Predict risk estimate of `X` as the probability of `X` to not be zero times the expected value of `X`:

        $$\text{score_sample(X)} = (1-P(X=0)) \cdot E[X]$$

        where:

        - $P(X=0)$ is calculated using the `.predict_proba()` method of the underlying classifier.
        - $E[X]$ is the regressor prediction on `X`.

        !!! info

            This method requires the underlying classifier to implement `.predict_proba()` method.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples,)
            The predicted risk.
        """

        check_is_fitted(self, ["n_features_in_", "classifier_", "regressor_"])
        X = validate_data(self, X=X, reset=False)

        non_zero_proba = self.classifier_.predict_proba(X)[:, 1]
        expected_impact = self.regressor_.predict(X)

        return non_zero_proba * expected_impact



from warnings import warn

import numpy as np
from scipy.optimize import minimize_scalar
from scipy.stats import gaussian_kde
from sklearn.base import BaseEstimator, OutlierMixin
from sklearn.mixture import GaussianMixture
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted
from sklearn_compat.utils.validation import validate_data


class GMMOutlierDetector(OutlierMixin, BaseEstimator):
    """The `GMMDetector` trains a Gaussian Mixture model on a dataset `X`. Once a density is trained we can evaluate the
    likelihood scores to see if it is deemed likely.

    By providing a `threshold` this model might then label outliers if their likelihood score is too low.

    !!! note
        The parameters other than `threshold` and `method` are an exact copy of the parameters in
        [sklearn.mixture.GaussianMixture]( https://scikit-learn.org/stable/modules/generated/sklearn.mixture.GaussianMixture.html).

    Parameters
    ----------
    threshold : float, default=0.99
        The limit at which the model thinks an outlier appears, must be between (0, 1).
    method : Literal["quantile", "stddev"], default="quantile"
        The method to use to apply the `threshold`.

        !!! info
            If you select `method="quantile"` then the threshold value represents the quantile value to start calling
            something an outlier.

            If you select `method="stddev"` then the threshold value represents the
            numbers of standard deviations before calling something an outlier.

    Attributes
    ----------
    gmm_ : GaussianMixture
        The trained Gaussian Mixture model.
    likelihood_threshold_ : float
        The threshold value used to determine if something is an outlier.

    Examples
    --------
    ```python
    import numpy as np
    from sklego.mixture import GMMOutlierDetector

    # Generate datset, it consists of two clusters
    np.random.seed(1)
    group0 = np.random.normal(0, 3, (10, 2))
    group1 = np.random.normal(2.5, 2, (5, 2))
    data = np.vstack([group0, group1])

    y = np.hstack([np.zeros((group0.shape[0],), dtype=int), np.ones((group1.shape[0],), dtype=int)])

    # Create and fit the GMMOutlierDetector model
    gmm = GMMOutlierDetector(threshold=0.9, n_components=2, random_state=1)
    gmm.fit(data, y)

    # Classify a new point as outlier or not
    p = np.array([[4.5, 0.5]])
    p_pred = gmm.predict(p) # predict the probabilities p belongs to each cluster
    print('The point is an outlier if the score is -1, inlier if the score is 1')
    ### The point is an outlier if the score is -1, inlier if the score is 1

    print(f'The score for this point is {p_pred}.')
    ### The score for this point is [-1].
    ```
    """

    _ALLOWED_METHODS = ("quantile", "stddev")

    def __init__(
        self,
        threshold=0.99,
        method="quantile",
        n_components=1,
        covariance_type="full",
        tol=1e-3,
        reg_covar=1e-6,
        max_iter=100,
        n_init=1,
        init_params="kmeans",
        weights_init=None,
        means_init=None,
        precisions_init=None,
        random_state=None,
        warm_start=False,
        verbose=0,
        verbose_interval=10,
    ):
        self.threshold = threshold
        self.method = method
        self.random_state = random_state
        self.n_components = n_components
        self.covariance_type = covariance_type
        self.tol = tol
        self.reg_covar = reg_covar
        self.max_iter = max_iter
        self.n_init = n_init
        self.init_params = init_params
        self.weights_init = weights_init
        self.means_init = means_init
        self.precisions_init = precisions_init
        self.random_state = random_state
        self.warm_start = warm_start
        self.verbose = verbose
        self.verbose_interval = verbose_interval

    def fit(self, X: np.ndarray, y=None) -> "GMMOutlierDetector":
        """Fit the `GMMOutlierDetector` model using `X`, `y` as training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features )
            The training data.
        y : array-like of shape (n_samples,)
            Ignored, present for compatibility.

        Returns
        -------
        self : GMMOutlierDetector
            The fitted estimator.

        Raises
        ------
        ValueError
            - If `method="quantile"` and `threshold` is not between (0, 1).
            - If `method="stddev"` and `threshold` is negative.
            - If `method` is not in `["quantile", "stddev"]`.
        """
        # GMM sometimes throws an error if you don't do this
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=True)
        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        if (self.method == "quantile") and ((self.threshold > 1) or (self.threshold < 0)):
            raise ValueError(f"Threshold {self.threshold} with method {self.method} needs to be 0 < threshold < 1")
        if (self.method == "stddev") and (self.threshold < 0):
            raise ValueError(f"Threshold {self.threshold} with method {self.method} needs to be 0 < threshold ")
        if self.method not in self._ALLOWED_METHODS:
            raise ValueError(f"Method not recognised. Method must be in {self._ALLOWED_METHODS}")

        self.gmm_ = GaussianMixture(
            n_components=self.n_components,
            covariance_type=self.covariance_type,
            tol=self.tol,
            reg_covar=self.reg_covar,
            max_iter=self.max_iter,
            n_init=self.n_init,
            init_params=self.init_params,
            weights_init=self.weights_init,
            means_init=self.means_init,
            precisions_init=self.precisions_init,
            random_state=self.random_state,
            warm_start=self.warm_start,
            verbose=self.verbose,
            verbose_interval=self.verbose_interval,
        )
        self.gmm_.fit(X)
        score_samples = self.gmm_.score_samples(X)

        if self.method == "quantile":
            self.likelihood_threshold_ = np.quantile(score_samples, 1 - self.threshold)

        if self.method == "stddev":
            density = gaussian_kde(score_samples)
            max_x_value = minimize_scalar(lambda x: -density(x)).x
            mean_likelihood = score_samples.mean()
            new_likelihoods = score_samples[score_samples < max_x_value]
            new_likelihoods_std = np.std(new_likelihoods - mean_likelihood)
            self.likelihood_threshold_ = mean_likelihood - (self.threshold * new_likelihoods_std)

        self.n_iter_ = self.gmm_.n_iter_
        self.n_features_in_ = X.shape[1]
        self.offset_ = self.likelihood_threshold_
        return self

    def score_samples(self, X):
        """Compute the log likelihood for each sample and return the negative value."""
        check_is_fitted(self, ["gmm_", "likelihood_threshold_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)

        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        return self.gmm_.score_samples(X)

    def decision_function(self, X):
        # We subtract self.offset_ to make 0 be the threshold value for being an outlier:
        return self.score_samples(X) - self.offset_

    def predict(self, X):
        """Predict if a point is an outlier or not using the fitted model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples,)
            The predicted data. 1 for inliers, -1 for outliers.
        """
        preds = (self.decision_function(X) >= 0).astype(int)
        preds[preds == 0] = -1
        return preds

    @property
    def allowed_methods(self):
        warn(
            "Please use `_ALLOWED_METHODS` instead of `allowed_methods`,"
            "`allowed_methods` will be deprecated in future versions",
            DeprecationWarning,
        )
        return self._ALLOWED_METHODS



import numpy as np
from scipy.special import softmax
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.mixture import BayesianGaussianMixture
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted
from sklearn_compat.utils.validation import validate_data


class BayesianGMMClassifier(ClassifierMixin, BaseEstimator):
    """The `BayesianGMMClassifier` trains a Gaussian Mixture Model for each class in `y` on a dataset `X`.
    Once a density is trained for each class we can evaluate the likelihood scores to see which class is more likely.

    !!! note
        All the parameters are an exact copy of those of
        [sklearn.mixture.BayesianGaussianMixture]( https://scikit-learn.org/stable/modules/generated/sklearn.mixture.BayesianGaussianMixture.html).


    Attributes
    ----------
    gmms_ : dict[int, BayesianGaussianMixture]
        A dictionary of Bayesian Gaussian Mixture Models, one for each class.
    classes_ : np.ndarray of shape (n_classes,)
        The classes seen during `fit`.

    Examples
    --------

    ```python
    import numpy as np
    from sklego.mixture import BayesianGMMClassifier

    # Generate datset
    np.random.seed(1)
    group0 = np.random.normal(0, 3, (1000, 2))
    group1 = np.random.normal(2.5, 2, (500, 2))
    data = np.vstack([group0, group1])

    y = np.hstack([np.zeros((group0.shape[0],), dtype=int), np.ones((group1.shape[0],), dtype=int)])
    # Create and fit the BayesianGMMClassifier model
    bgmm = BayesianGMMClassifier(n_components=2, random_state=1)
    bgmm.fit(data, y)

    # Classify the train dataset into two clusters (n_components=2)
    labels = bgmm.predict(data)

    # Classify a new point into one of two clusters
    p = np.array([[1.5, 0.5]])
    p_prob = bgmm.predict_proba(p) # predict the probabilities p belongs to each cluster
    print(f'Probability point p belongs to group1 is {p_prob[0,0]:.2f}')
    ### Probability point p belongs to group1 is 0.38
    print(f'Probability point p belongs to group2 is {p_prob[0,1]:.2f}')
    ### Probability point p belongs to group2 is 0.62

    print(f'It is more probable that point p belongs to group{np.argmax(p_prob)}')
    ### It is more probable that point p belongs to group1
    ```
    """

    def __init__(
        self,
        n_components=1,
        covariance_type="full",
        tol=0.001,
        reg_covar=1e-06,
        max_iter=100,
        n_init=1,
        init_params="kmeans",
        weight_concentration_prior_type="dirichlet_process",
        weight_concentration_prior=None,
        mean_precision_prior=None,
        mean_prior=None,
        degrees_of_freedom_prior=None,
        covariance_prior=None,
        random_state=None,
        warm_start=False,
        verbose=0,
        verbose_interval=10,
    ):
        self.n_components = n_components
        self.covariance_type = covariance_type
        self.tol = tol
        self.reg_covar = reg_covar
        self.max_iter = max_iter
        self.n_init = n_init
        self.init_params = init_params
        self.weight_concentration_prior_type = weight_concentration_prior_type
        self.weight_concentration_prior = weight_concentration_prior
        self.mean_precision_prior = mean_precision_prior
        self.mean_prior = mean_prior
        self.degrees_of_freedom_prior = degrees_of_freedom_prior
        self.covariance_prior = covariance_prior
        self.random_state = random_state
        self.warm_start = warm_start
        self.verbose = verbose
        self.verbose_interval = verbose_interval

    def fit(self, X: np.ndarray, y: np.ndarray) -> "BayesianGMMClassifier":
        """Fit the `BayesianGMMClassifier` model using `X`, `y` as training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features )
            The training data.
        y : array-like of shape (n_samples,)
            The target values.

        Returns
        -------
        self : BayesianGMMClassifier
            The fitted estimator.
        """
        X, y = validate_data(self, X=X, y=y, dtype=FLOAT_DTYPES, reset=True)
        if X.ndim == 1:
            X = np.expand_dims(X, 1)

        self.gmms_ = {}
        self.classes_ = unique_labels(y)
        for c in self.classes_:
            subset_x, subset_y = X[y == c], y[y == c]
            mixture = BayesianGaussianMixture(
                n_components=self.n_components,
                covariance_type=self.covariance_type,
                tol=self.tol,
                reg_covar=self.reg_covar,
                max_iter=self.max_iter,
                n_init=self.n_init,
                init_params=self.init_params,
                weight_concentration_prior_type=self.weight_concentration_prior_type,
                weight_concentration_prior=self.weight_concentration_prior,
                mean_precision_prior=self.mean_precision_prior,
                mean_prior=self.mean_prior,
                degrees_of_freedom_prior=self.degrees_of_freedom_prior,
                covariance_prior=self.covariance_prior,
                random_state=self.random_state,
                warm_start=self.warm_start,
                verbose=self.verbose,
                verbose_interval=self.verbose_interval,
            )
            self.gmms_[c] = mixture.fit(subset_x, subset_y)

        self.n_iter_ = sum(mixture.n_iter_ for mixture in self.gmms_.values())

        return self

    def predict(self, X):
        """Predict labels for `X` using fitted estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples,)
            The predicted data.
        """
        check_is_fitted(self, ["gmms_", "classes_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)

        return self.classes_[self.predict_proba(X).argmax(axis=1)]

    def predict_proba(self, X):
        """Predict probabilities for `X` using fitted estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples, n_classes)
            The predicted probabilities.
        """
        check_is_fitted(self, ["gmms_", "classes_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)

        res = np.zeros((X.shape[0], self.classes_.shape[0]))
        for idx, c in enumerate(self.classes_):
            res[:, idx] = self.gmms_[c].score_samples(X)
        return softmax(res, axis=1)



import numpy as np
from sklearn import clone
from sklearn.base import BaseEstimator, ClassifierMixin, MetaEstimatorMixin
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import normalize
from sklearn.utils.validation import FLOAT_DTYPES, check_is_fitted
from sklearn_compat.utils.validation import validate_data


class SubjectiveClassifier(ClassifierMixin, MetaEstimatorMixin, BaseEstimator):
    """Corrects predictions of the inner classifier by taking into account a (subjective) prior distribution of the
    classes.

    This can be useful when there is a difference in class distribution between the training data set and the real
    world. Using the confusion matrix of the inner classifier and the prior, the posterior probability for a class,
    given the prediction of the inner classifier, can be computed.

    The background for this posterior estimation is given in
    [this article](https://lucdemortier.github.io/articles/16/PerformanceMetrics).

    Based on the `evidence` attribute, this meta estimator's predictions are based on simple weighing of the inner
    estimator's `predict_proba()` results, the posterior probabilities based on the confusion matrix, or a combination
    of the two approaches.

    Parameters
    ----------
    estimator : scikit-learn compatible classifier
        Classifier that will be wrapped with SubjectiveClassifier. It should implement `predict_proba` method.
    prior : dict[int, float]
        A dictionary mapping `class -> frequency` representing the prior (a.k.a. subjective real-world) class
        distribution. The class frequencies should sum to 1.
    evidence : Literal["predict_proba", "confusion_matrix", "both"], default="both"
        A string indicating which evidence should be used to correct the inner estimator's predictions.

        - If `"both"`  the inner estimator's `predict_proba()` results are multiplied by the posterior probabilities.
        - If `"predict_proba"`, the inner estimator's `predict_proba()` results are multiplied by the prior
            distribution.
        - If `"confusion_matrix"`, the inner estimator's discrete predictions are converted to posterior probabilities
            using the prior and the inner estimator's confusion matrix (obtained from the train data used in `fit()`).

    Attributes
    ----------
    estimator_ : scikit-learn compatible classifier
        The fitted classifier.
    classes_ : array-like, shape=(n_classes,)
        The classes labels.
    posterior_matrix_ : array-like, shape=(n_classes, n_classes)
        The posterior probabilities for each class, given the prediction of the inner classifier.

    Example
    -------
    ```py
    import numpy as np
    from sklearn.linear_model import LogisticRegression
    from sklego.meta import SubjectiveClassifier

    np.random.seed(0)
    n1, n2 = 50, 100
    X = np.concatenate([np.random.normal(0, 1, (n1, 2)), np.random.normal(2, 1, (n2, 2))], axis=0)
    y = np.concatenate([np.zeros((n1, 1)), np.ones((n2, 1))], axis=0).reshape(-1)

    prior = {0: 0.2, 1: 0.8}

    logistic_regressor = LogisticRegression()
    subjective_classifier = SubjectiveClassifier(logistic_regressor, prior, evidence="confusion_matrix")
    subjective_classifier.fit(X, y)

    # Classify new datapoint

    dp = np.random.normal(0, 1, (1,2)) # comes from the same distribution as datappints in class 0
    pred_prob = subjective_classifier.predict_proba(dp)
    preds = subjective_classifier.predict(dp)
    print(f"Datapoint {dp[0]} with predicted probabilities: {pred_prob[0]}, is classified as belonging in class {preds[0]}")
    ### Datapoint [-1.30652685  1.65813068] with predicted probabilities: [0.91836735 0.08163265], is classified as belonging in class 0.0
    ```
    """

    _ALLOWED_EVIDENCE = ("predict_proba", "confusion_matrix", "both")
    _required_parameters = ["estimator", "prior"]

    def __init__(self, estimator, prior, evidence="both"):
        self.estimator = estimator
        self.prior = prior
        self.evidence = evidence

    def _likelihood(self, predicted_class, given_class, cfm):
        return cfm[given_class, predicted_class] / cfm[given_class, :].sum()

    def _evidence(self, predicted_class, cfm):
        return sum(
            [
                self._likelihood(predicted_class, given_class, cfm) * self.prior[self.classes_[given_class]]
                for given_class in range(cfm.shape[0])
            ]
        )

    def _posterior(self, y, y_hat, cfm):
        y_hat_evidence = self._evidence(y_hat, cfm)
        return (
            (self._likelihood(y_hat, y, cfm) * self.prior[self.classes_[y]] / y_hat_evidence)
            if y_hat_evidence > 0
            else self.prior[y]  # in case confusion matrix has all-zero column for y_hat
        )

    def fit(self, X, y):
        """Fit the inner classfier using `X` and `y` as training data by fitting the underlying estimator and computing
        the posterior probabilities.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features )
            The training data.
        y : array-like of shape (n_samples,)
            The target values.

        Returns
        -------
        self : SubjectiveClassifier
            The fitted estimator.

        Raises
        ------
        ValueError
            - If `estimator` is not a classifier.
            - If `y` contains classes that are not specified in the `prior`
            - If `prior` is not a valid probability distribution (i.e. does not sum to 1).
            - If `evidence` is not one of "predict_proba", "confusion_matrix", or "both".
        """
        if not isinstance(self.estimator, ClassifierMixin):
            raise ValueError(
                "Invalid inner estimator: the SubjectiveClassifier meta model only works on classification models"
            )

        if not np.isclose(sum(self.prior.values()), 1):
            raise ValueError("Invalid prior: the prior probabilities of all classes should sum to 1")

        if self.evidence not in self._ALLOWED_EVIDENCE:
            raise ValueError(f"Invalid evidence: the provided evidence should be one of {self._ALLOWED_EVIDENCE}")

        X, y = validate_data(self, X=X, y=y, dtype=FLOAT_DTYPES, reset=True)

        if set(y) - set(self.prior.keys()):
            raise ValueError(
                f"Training data is inconsistent with prior: no prior defined for classes "
                f"{set(y) - set(self.prior.keys())}"
            )
        self.estimator_ = clone(self.estimator).fit(X, y)
        cfm = confusion_matrix(y, self.estimator_.predict(X))
        self.posterior_matrix_ = np.array(
            [[self._posterior(y, y_hat, cfm) for y_hat in range(cfm.shape[0])] for y in range(cfm.shape[0])]
        )
        return self

    @staticmethod
    def _weighted_proba(weights, y_hat_probas):
        return normalize(weights * y_hat_probas, norm="l1")

    @staticmethod
    def _to_discrete(y_hat_probas):
        y_hat_discrete = np.zeros(y_hat_probas.shape)
        y_hat_discrete[np.arange(y_hat_probas.shape[0]), y_hat_probas.argmax(axis=1)] = 1
        return y_hat_discrete

    def predict_proba(self, X):
        """Predict probability distribution of the class, based on the provided data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples, n_classes)
            The predicted probabilities.
        """
        check_is_fitted(self, ["posterior_matrix_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)

        y_hats = self.estimator_.predict_proba(X)  # these are ignorant of the prior

        if self.evidence == "predict_proba":
            prior_weights = np.array([self.prior[klass] for klass in self.classes_])
            return self._weighted_proba(prior_weights, y_hats)
        else:
            posterior_probas = self._to_discrete(y_hats) @ self.posterior_matrix_.T
            return self._weighted_proba(posterior_probas, y_hats) if self.evidence == "both" else posterior_probas

    def predict(self, X):
        """Predict target values for `X` using fitted estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data to predict.

        Returns
        -------
        array-like of shape (n_samples, )
            The predicted class.
        """
        check_is_fitted(self, ["posterior_matrix_"])
        X = validate_data(self, X=X, dtype=FLOAT_DTYPES, reset=False)

        return self.classes_[self.predict_proba(X).argmax(axis=1)]

    @property
    def classes_(self):
        """Alias for `.classes_` attribute of the underlying estimator."""
        return self.estimator_.classes_



from copy import deepcopy
from typing import List, Union

import narwhals.stable.v1 as nw
import numpy as np
import pandas as pd
from sklearn import clone
from sklearn.base import BaseEstimator, ClassifierMixin, MetaEstimatorMixin, RegressorMixin, is_classifier, is_regressor
from sklearn.utils.metaestimators import available_if
from sklearn.utils.validation import check_is_fitted

from sklego.common import as_list, expanding_list
from sklego.meta._grouped_utils import parse_X_y
from sklego.meta._shrinkage_utils import (
    ShrinkageMixin,
    constant_shrinkage,
    equal_shrinkage,
    min_n_obs_shrinkage,
    relative_shrinkage,
)


class GroupedPredictor(ShrinkageMixin, MetaEstimatorMixin, BaseEstimator):
    """`GroupedPredictor` is a meta-estimator that fits a separate estimator for each group in the input data.

    The input data is split into a group and a value part: for each unique combination of the group columns, a separate
    estimator is fitted to the corresponding value rows. The group columns are specified by the `groups` parameter.

    If `use_global_model=True` a fallback estimator will be fitted on the entire dataset in case a group is not found
    during `.predict()`.

    If `shrinkage` is not `None`, the predictions of the group-level models are combined using a shrinkage method. The
    shrinkage method can be one of the predefined methods `"constant"`, `"equal"`, `"min_n_obs"`, `"relative"` or a
    custom shrinkage function. The shrinkage method is specified by the `shrinkage` parameter.

    !!! warning "Shrinkage"
        Shrinkage is only available for regression models.

    Parameters
    ----------
    estimator : scikit-learn compatible estimator/pipeline
        The estimator/pipeline to be applied per group.
    groups : int | str | List[int] | List[str]
        The column(s) of the array/dataframe to select as a grouping parameter set.
    shrinkage : Literal["constant", "equal", "min_n_obs", "relative"] | Callable | None, default=None
        How to perform shrinkage:

        - `None`: No shrinkage (default)
        - `"constant"`: the augmented prediction for each level is the weighted average between its prediction and the
            augmented prediction for its parent.
        - `"equal"`: each group is weighed equally.
        - `"min_n_obs"`: use only the smallest group with a certain amount of observations.
        - `"relative"`: weigh each group according to its size.
        - `Callable`: a function that takes a list of group lengths and returns an array of the same size with the
            weights for each group.
    use_global_model : bool, default=True

        - With shrinkage: whether to have a model over the entire input as first group
        - Without shrinkage: whether or not to fall back to a general model in case the group parameter is not found
            during `.predict()`
    check_X : bool, default=True
        Whether to validate `X` to be non-empty 2D array of finite values and attempt to cast `X` to float.
        If disabled, the model/pipeline is expected to handle e.g. missing, non-numeric, or non-finite values.
    **shrinkage_kwargs : dict
        Keyword arguments to the shrinkage function

    Attributes
    ----------
    estimators_ : dict
        A dictionary with the fitted estimators per group
    groups_ : list
        A list of all the groups that were found during fitting
    fallback_ : estimator
        A fallback estimator that is used when `use_global_model=True` and a group is not found during `.predict()`
    shrinkage_function_ : callable
        The shrinkage function that is used to calculate the shrinkage factors
    shrinkage_factors_ : dict
        A dictionary with the shrinkage factors per group

    Example
    -------
    ```py
    import pandas as pd
    from sklearn.dummy import DummyRegressor
    from sklego.meta import GroupedPredictor

    results_df = pd.DataFrame(
        {
            "Grade": ["11", "11", "11", "11", "12", "12", "12", "12"],
            "Course": ["Algebra", "Algebra", "English", "English", "Algebra", "Algebra", "English", "English"],
            "Name": ["Mary", "Helen", "Mary", "Helen", "Mary", "Helen", "Mary", "Helen"],
            "Result": [100, 94, 88, 92, 96, 98, 90, 90],
        }
    )

    groups = ["Grade", "Name"]
    target = "Result"

    # We will use the DummyRegressor() to calculate the mean of each grouping
    grouped_pred = GroupedPredictor(DummyRegressor(), groups)
    grouped_pred.fit(results_df[groups], results_df[target])

    # What is the average for Mary in Grade 12?
    pred_df = pd.DataFrame(
        {
            "Grade": ["12"],
            "Name": ["Mary"],
        }
    )

    # Predicts the mean result of each student in each grade
    pred = grouped_pred.predict(pred_df)

    print(f"The average result of {pred_df["Name"][0]} in Grade {pred_df["Grade"][0]} was {pred}")
    ### The average result of Mary in Grade 12 was 93.0
    ```
    """

    # Number of features in value df can be 0, e.g. for dummy models
    _check_kwargs = {"ensure_min_features": 0, "accept_large_sparse": False}
    _global_col_name = "a-column-that-is-constant-for-all-data"
    _global_col_value = "global"

    _ALLOWED_SHRINKAGE = {
        "constant": constant_shrinkage,
        "relative": relative_shrinkage,
        "min_n_obs": min_n_obs_shrinkage,
        "equal": equal_shrinkage,
    }

    _required_parameters = ["estimator", "groups"]

    def __init__(
        self,
        estimator,
        groups,
        shrinkage=None,
        use_global_model=True,
        check_X=True,
        shrinkage_kwargs=None,
    ):
        self.estimator = estimator
        self.groups = groups
        self.shrinkage = shrinkage
        self.use_global_model = use_global_model
        self.shrinkage_kwargs = shrinkage_kwargs
        self.check_X = check_X

    def __fit_single_group(self, group, X, y=None):
        """Fit estimator to the given group."""
        try:
            return clone(self.estimator).fit(X, y)
        except Exception as e:
            raise type(e)(f"Exception for group {group}: {e}")

    def __fit_grouped_estimator(
        self, frame: nw.DataFrame, y: Union[np.ndarray, None] = None, columns: Union[List[int], List[str], None] = None
    ):
        """Fit an estimator to each group"""

        if columns is None:
            columns = self._groups

        to_drop = list(set(["__sklego_target__", *columns, *as_list(self.groups)]))
        grouped_estimators = {
            # Fit a clone of the estimators to each group
            (group_name[0] if len(group_name) == 1 else group_name): self.__fit_single_group(
                group=(group_name[0] if len(group_name) == 1 else group_name),
                X=nw.to_native(X_grp.drop(to_drop)),
                y=(X_grp.select("__sklego_target__").to_numpy().reshape(-1) if y is not None else None),
            )
            for group_name, X_grp in frame.group_by(columns)
        }

        return grouped_estimators

    def __fit_shrinkage_groups(self, frame, y):
        estimators = {}

        for grouping_colnames in self.group_colnames_hierarchical_:
            # Fit a grouped estimator to each (sub)group hierarchically
            estimators.update(self.__fit_grouped_estimator(frame, y, columns=grouping_colnames))

        return estimators

    def __add_shrinkage_column(self, frame, groups=None):
        """Add global group as first column if needed for shrinkage"""

        if self.shrinkage is not None and self.use_global_model:
            n_samples = frame.shape[0]

            frame = frame.select(
                nw.from_dict(
                    data={self._global_col_name: np.full(shape=n_samples, fill_value=self._global_col_value)},
                    native_namespace=nw.get_native_namespace(frame),
                )[self._global_col_name],
                nw.all(),
            )
            groups = [self._global_col_name] if groups is None else [self._global_col_name, *groups]

        return frame, groups

    def fit(self, X, y=None):
        """Fit one estimator for each group of training data `X` and `y`.

        Will also learn the groups that exist within the dataset.

        If `use_global_model=True` a fallback estimator will be fitted on the entire dataset in case a group is not
        found during `.predict()`.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,), default=None
            Target values.

        Returns
        -------
        self : GroupedPredictor
            The fitted estimator.
        """
        if self.shrinkage is not None and not is_regressor(self.estimator):
            raise ValueError("Shrinkage is only available for regression models")

        _group_cols = as_list(deepcopy(self.groups)) if self.groups is not None else None

        if (
            self.shrinkage is not None
            and _group_cols is not None
            and len(_group_cols) == 1
            and not self.use_global_model
        ):
            raise ValueError("Shrinkage is not null, but found a total of 1 groups")

        X = nw.from_native(X, strict=False, eager_only=True)

        frame = parse_X_y(X, y, _group_cols, check_X=self.check_X, **self._check_kwargs)
        frame, _group_cols = self.__add_shrinkage_column(frame, _group_cols)
        self.n_features_in_ = frame.shape[1] - 1
        self.n_fitted_levels_ = 1 + self.use_global_model

        self.shrinkage_function_ = self._set_shrinkage_function()

        # List of all hierarchical subsets of columns
        self.group_colnames_hierarchical_ = expanding_list(_group_cols, list)
        self.fallback_ = None

        if self.shrinkage is None and self.use_global_model:
            X_ = nw.to_native(frame.drop([*_group_cols, "__sklego_target__"]))
            y_ = nw.to_native(frame["__sklego_target__"])

            self.fallback_ = clone(self.estimator).fit(X_, y_)

        if self.shrinkage is not None:
            self.estimators_ = self.__fit_shrinkage_groups(frame, y)
        else:
            self.estimators_ = self.__fit_grouped_estimator(frame, y, columns=_group_cols)

        self.groups_ = as_list(self.estimators_.keys())

        if self.shrinkage is not None:
            _groups = (
                [self._global_col_name, *as_list(deepcopy(self.groups))]
                if self.use_global_model
                else as_list(deepcopy(self.groups))
            )

            self.shrinkage_factors_ = self._fit_shrinkage_factors(frame, groups=_groups, most_granular_only=True)
            self.shrinkage_factors_ = {(k[0] if len(k) == 1 else k): v for k, v in self.shrinkage_factors_.items()}

        return self

    def __predict_shrinkage_groups(self, frame, method="predict", groups=None):
        """Make predictions for all shrinkage groups"""
        # DataFrame with predictions for each hierarchy level, per row. Missing groups errors are thrown here.
        hierarchical_predictions = pd.concat(
            [
                pd.Series(self.__predict_groups(frame, method=method, groups=level_columns))
                for level_columns in self.group_colnames_hierarchical_
            ],
            axis=1,
        )

        # This is a Series with values the tuples of hierarchical grouping
        prediction_groups = pd.Series([tuple(_) for _ in frame.select(groups).to_pandas().itertuples(index=False)])

        # This is a Series of arrays
        shrinkage_factors = prediction_groups.map(self.shrinkage_factors_)

        # Convert the Series of arrays it to a DataFrame
        shrinkage_factors = pd.DataFrame.from_dict(shrinkage_factors.to_dict()).T

        return (hierarchical_predictions * shrinkage_factors).sum(axis=1)

    def __predict_single_group(self, group, X, method="predict"):
        """Predict a single group by getting its estimator from the fitted dict"""

        try:
            group_predictor = self.estimators_[group]
        except KeyError:
            if self.fallback_:
                group_predictor = self.fallback_
            else:
                raise ValueError(f"Found new group {group} during predict with use_global_model = False")

        is_predict_proba = is_classifier(group_predictor) and method == "predict_proba"
        # Ensure to provide pd.DataFrame with the correct label name
        extra_kwargs = {"columns": group_predictor.classes_} if is_predict_proba else {}

        # getattr(group_predictor, method) returns the predict method of the fitted model
        # if the method argument is "predict" and the predict_proba method if method argument is "predict_proba"
        return pd.DataFrame(getattr(group_predictor, method)(X), **extra_kwargs)

    def __predict_groups(self, frame: nw.DataFrame, method="predict", groups=None):
        """Predict for all groups"""

        n_samples = frame.shape[0]
        frame = frame.with_columns(__sklego_index__=np.arange(n_samples))
        return (
            pd.concat(
                [
                    self.__predict_single_group(
                        (group_value[0] if len(group_value) == 1 else group_value),
                        nw.to_native(X_grp.drop(["__sklego_index__", *groups, *as_list(self.groups)])),
                        method=method,
                    ).set_index(X_grp["__sklego_index__"].to_numpy().reshape(-1).astype(int))
                    for group_value, X_grp in frame.group_by(groups)
                ],
                axis=0,
            )
            .fillna(0)
            .sort_index()
            .to_numpy()
            .squeeze()
        )

    def predict(self, X):
        """Predict target values on new data `X` by predicting on each group. If a group is not found during
        `.predict()` and `use_global_model=True` the fallback estimator will be used. If `use_global_model=False` a
        `ValueError` will be raised.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data to predict.

        Returns
        -------
        array-like of shape (n_samples,)
            Predicted target values.
        """
        check_is_fitted(self, ["estimators_", "groups_", "fallback_"])

        _group_cols = as_list(deepcopy(self.groups)) if self.groups is not None else None
        X = nw.from_native(X, strict=False, eager_only=True)
        frame = parse_X_y(X, y=None, groups=_group_cols, check_X=self.check_X, **self._check_kwargs).drop(
            "__sklego_target__"
        )
        frame, _group_cols = self.__add_shrinkage_column(frame, _group_cols)

        if self.shrinkage is None:
            return self.__predict_groups(frame, method="predict", groups=_group_cols)
        else:
            return self.__predict_shrinkage_groups(frame, method="predict", groups=_group_cols)

    # This ensures that the meta-estimator only has the predict_proba method if the estimator has it
    @available_if(lambda self: hasattr(self.estimator, "predict_proba"))
    def predict_proba(self, X):
        """Predict probabilities on new data `X`.

        !!! warning
            Available only if the underlying estimator implements `.predict_proba()` method.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data to predict.

        Returns
        -------
        array-like of shape (n_samples, n_classes)
            Predicted probabilities per class.
        """
        check_is_fitted(self, ["estimators_", "groups_", "fallback_"])

        _group_cols = as_list(deepcopy(self.groups)) if self.groups is not None else None
        X = nw.from_native(X, strict=False, eager_only=True)
        frame = parse_X_y(X, y=None, groups=_group_cols, check_X=self.check_X, **self._check_kwargs).drop(
            "__sklego_target__"
        )
        frame, _group_cols = self.__add_shrinkage_column(frame, _group_cols)

        if self.shrinkage is None:
            return self.__predict_groups(frame, method="predict_proba", groups=_group_cols)
        else:
            return self.__predict_shrinkage_groups(frame, method="predict_proba", groups=_group_cols)

    # This ensures that the meta-estimator only has the predict_proba method if the estimator has it
    @available_if(lambda self: hasattr(self.estimator, "decision_function"))
    def decision_function(self, X):
        """Predict confidence scores for samples in `X`.

        !!! warning
            Available only if the underlying estimator implements `.decision_function()` method.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data to predict.

        Returns
        -------
        array-like of shape (n_samples,) or (n_samples, n_classes)
            Confidence scores per (n_samples, n_classes) combination.
            In the binary case, confidence score for self.classes_[1] where > 0 means this class would be
            predicted.
        """
        check_is_fitted(self, ["estimators_", "groups_", "fallback_"])

        _group_cols = as_list(deepcopy(self.groups)) if self.groups is not None else None
        X = nw.from_native(X, strict=False, eager_only=True)

        frame = parse_X_y(X, y=None, groups=_group_cols, check_X=self.check_X, **self._check_kwargs).drop(
            "__sklego_target__"
        )
        frame, _group_cols = self.__add_shrinkage_column(frame, _group_cols)

        if self.shrinkage is None:
            return self.__predict_groups(frame, method="decision_function", groups=_group_cols)
        else:
            return self.__predict_shrinkage_groups(frame, method="decision_function", groups=_group_cols)

    @property
    def _estimator_type(self):
        """Computes `_estimator_type` dynamically from the wrapped model."""
        return self.estimator._estimator_type

    def _more_tags(self):
        return {"allow_nan": True}

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        return tags


class GroupedRegressor(RegressorMixin, GroupedPredictor):
    """`GroupedRegressor` is a meta-estimator that fits a separate regressor for each group in the input data.

    Its spec is the same as [`GroupedPredictor`][sklego.meta.grouped_predictor.GroupedPredictor] but it is available
    only for regression models.

    !!! info "New in version 0.8.0"
    """

    def fit(self, X, y):
        """Fit one regressor for each group of training data `X` and `y`.

        Will also learn the groups that exist within the training dataset.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        self : GroupedRegressor
            The fitted regressor.

        Raises
        -------
        ValueError
            If the supplied estimator is not a regressor.

        """
        if not is_regressor(self.estimator):
            raise ValueError("GroupedRegressor is only available for regression models")

        return super().fit(X, y)


class GroupedClassifier(ClassifierMixin, GroupedPredictor):
    """`GroupedClassifier` is a meta-estimator that fits a separate classifier for each group in the input data.

    Its equivalent to [`GroupedPredictor`][sklego.meta.grouped_predictor.GroupedPredictor] with `shrinkage=None`
    but it is available only for classification models.

    !!! info "New in version 0.8.0"
    """

    def __init__(
        self,
        estimator,
        groups,
        use_global_model=True,
        check_X=True,
        **shrinkage_kwargs,
    ):
        super().__init__(
            estimator=estimator,
            groups=groups,
            shrinkage=None,
            use_global_model=use_global_model,
            check_X=check_X,
        )

    def fit(self, X, y):
        """Fit one classifier for each group of training data `X` and `y`.

        Will also learn the groups that exist within the training dataset.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data.
        y : array-like of shape (n_samples,)
            Target values.

        Returns
        -------
        self : GroupedClassifier
            The fitted classifier.

        Raises
        -------
        ValueError
            If the supplied estimator is not a classifier.
        """

        if not is_classifier(self.estimator):
            raise ValueError("GroupedClassifier is only available for classification models")
        self.classes_ = np.unique(y)
        return super().fit(X, y)
